import logging

import click

logging.basicConfig(format='%(asctime)s [%(filename)s:%(lineno)d] %(message)s',
                    datefmt='%m/%d/%Y %I:%M:%S %p')  # , level=logging.DEBUG)

CONTEXT_SETTINGS = dict(
    # Support -h as a shortcut for --help
    help_option_names=['-h', '--help'],
)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.pass_context
def main(ctx):
    """

    :param ctx:
    """
    ctx.obj = dict()



""" The plugins package is responsible for loading model helper functions supporting different ML frameworks. The :class:`fedn.utils.plugins.helperbase.HelperBase` is
an abstract class which user can implement their own helper functions to support different ML frameworks. """
# flake8: noqa



from concurrent import futures

import grpc

import fedn.common.net.grpc.fedn_pb2_grpc as rpc


class Server:
    """

    """

    def __init__(self, servicer, modelservicer, config):

        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=350))
        self.certificate = None

        if isinstance(servicer, rpc.CombinerServicer):
            rpc.add_CombinerServicer_to_server(servicer, self.server)
        if isinstance(servicer, rpc.ConnectorServicer):
            rpc.add_ConnectorServicer_to_server(servicer, self.server)
        if isinstance(servicer, rpc.ReducerServicer):
            rpc.add_ReducerServicer_to_server(servicer, self.server)
        if isinstance(modelservicer, rpc.ModelServiceServicer):
            rpc.add_ModelServiceServicer_to_server(modelservicer, self.server)
        if isinstance(servicer, rpc.CombinerServicer):
            rpc.add_ControlServicer_to_server(servicer, self.server)

        if config['secure']:
            print("Creating secure gRPCS server using certificate: {config['certificate']}", flush=True)
            server_credentials = grpc.ssl_server_credentials(
                ((config['key'], config['certificate'],),))
            self.server.add_secure_port(
                '[::]:' + str(config['port']), server_credentials)
        else:
            print("Creating insecure gRPC server", flush=True)
            self.server.add_insecure_port('[::]:' + str(config['port']))

    def start(self):
        """

        """
        print("Server started", flush=True)
        self.server.start()

    def stop(self):
        """

        """
        self.server.stop(0)



import abc


class Repository(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def set_artifact(self, instance_name, instance):
        """

        :param instance_name:
        :param instance:
        """
        raise NotImplementedError("Must be implemented by subclass")

    @abc.abstractmethod
    def get_artifact(self, instance_name):
        """

        :param instance_name:
        """
        raise NotImplementedError("Must be implemented by subclass")



import io
import logging

from minio import Minio
from minio.error import InvalidResponseError
from urllib3.poolmanager import PoolManager

from .base import Repository

logger = logging.getLogger(__name__)


class MINIORepository(Repository):
    """

    """
    client = None

    def __init__(self, config):
        super().__init__()
        try:
            access_key = config['storage_access_key']
        except Exception:
            access_key = 'minio'
        try:
            secret_key = config['storage_secret_key']
        except Exception:
            secret_key = 'minio123'
        try:
            self.bucket = config['storage_bucket']
        except Exception:
            self.bucket = 'fedn-models'
        try:
            self.context_bucket = config['context_bucket']
        except Exception:
            self.bucket = 'fedn-context'
        try:
            self.secure_mode = bool(config['storage_secure_mode'])
        except Exception:
            self.secure_mode = False

        if not self.secure_mode:
            print(
                "\n\n\nWARNING : S3/MINIO RUNNING IN **INSECURE** MODE! THIS IS NOT FOR PRODUCTION!\n\n\n")

        if self.secure_mode:
            manager = PoolManager(
                num_pools=100, cert_reqs='CERT_NONE', assert_hostname=False)
            self.client = Minio("{0}:{1}".format(config['storage_hostname'], config['storage_port']),
                                access_key=access_key,
                                secret_key=secret_key,
                                secure=self.secure_mode, http_client=manager)
        else:
            self.client = Minio("{0}:{1}".format(config['storage_hostname'], config['storage_port']),
                                access_key=access_key,
                                secret_key=secret_key,
                                secure=self.secure_mode)

        # TODO: generalize
        self.context_bucket = 'fedn-context'
        self.create_bucket(self.context_bucket)
        self.create_bucket(self.bucket)

    def create_bucket(self, bucket_name):
        """ Create a new bucket. If bucket exists, do nothing.

        :param bucket_name: The name of the bucket
        :type bucket_name: str
        """
        found = self.client.bucket_exists(bucket_name)

        if not found:
            try:
                self.client.make_bucket(bucket_name)
            except InvalidResponseError:
                raise

    def set_artifact(self, instance_name, instance, is_file=False, bucket=''):
        """ Instance must be a byte-like object. """
        if bucket == '':
            bucket = self.bucket
        if is_file:
            self.client.fput_object(bucket, instance_name, instance)
        else:
            try:
                self.client.put_object(
                    bucket, instance_name, io.BytesIO(instance), len(instance))
            except Exception as e:
                raise Exception("Could not load data into bytes {}".format(e))

        return True

    def get_artifact(self, instance_name, bucket=''):
        """

        :param instance_name:
        :param bucket:
        :return:
        """
        if bucket == '':
            bucket = self.bucket

        try:
            data = self.client.get_object(bucket, instance_name)
            return data.read()
        except Exception as e:
            raise Exception("Could not fetch data from bucket, {}".format(e))

    def get_artifact_stream(self, instance_name):
        """

        :param instance_name:
        :return:
        """
        try:
            data = self.client.get_object(self.bucket, instance_name)
            return data
        except Exception as e:
            raise Exception("Could not fetch data from bucket, {}".format(e))

    def list_artifacts(self):
        """

        :return:
        """
        objects_to_delete = []
        try:
            objs = self.client.list_objects(self.bucket)
            for obj in objs:
                print(obj.object_name)
                objects_to_delete.append(obj.object_name)
        except Exception:
            raise Exception(
                "Could not list models in bucket {}".format(self.bucket))
        return objects_to_delete

    def delete_artifact(self, instance_name, bucket=[]):
        """

        :param instance_name:
        :param bucket:
        """
        if not bucket:
            bucket = self.bucket

        try:
            self.client.remove_object(bucket, instance_name)
        except InvalidResponseError as err:
            print(err)
            print('Could not delete artifact: {}'.format(instance_name))

    def delete_objects(self):
        """

        """
        objects_to_delete = self.list_artifacts()
        try:
            # Remove list of objects.
            errors = self.client.remove_objects(
                self.bucket, objects_to_delete
            )
            for del_err in errors:
                print("Deletion Error: {}".format(del_err))
        except Exception:
            print('Could not delete objects: {}'.format(objects_to_delete))



import os
import tempfile
from io import BytesIO

import fedn.common.net.grpc.fedn_pb2 as fedn
import fedn.common.net.grpc.fedn_pb2_grpc as rpc
from fedn.common.storage.models.tempmodelstorage import TempModelStorage

CHUNK_SIZE = 1024 * 1024


class ModelService(rpc.ModelServiceServicer):
    """ Service for handling download and upload of models to the server.

    """

    def __init__(self):
        self.models = TempModelStorage()

    def exist(self, model_id):
        """ Check if a model exists on the server.

        :param model_id: The model id.
        :return: True if the model exists, else False.
        """
        return self.models.exist(model_id)

    def get_tmp_path(self):
        """ Return a temporary output path compatible with save_model, load_model. """
        fd, path = tempfile.mkstemp()
        os.close(fd)
        return path

    def load_model_from_BytesIO(self, model_bytesio, helper):
        """ Load a model from a BytesIO object.

        :param model_bytesio: A BytesIO object containing the model.
        :type model_bytesio: :class:`io.BytesIO`
        :param helper: The helper object for the model.
        :type helper: :class:`fedn.utils.helperbase.HelperBase`
        :return: The model object.
        :rtype: return type of helper.load
        """
        path = self.get_tmp_path()
        with open(path, 'wb') as fh:
            fh.write(model_bytesio)
            fh.flush()
        model = helper.load(path)
        os.unlink(path)
        return model

    def serialize_model_to_BytesIO(self, model, helper):
        """ Serialize a model to a BytesIO object.

        :param model: The model object.
        :type model: return type of helper.load
        :param helper: The helper object for the model.
        :type helper: :class:`fedn.utils.helperbase.HelperBase`
        :return: A BytesIO object containing the model.
        :rtype: :class:`io.BytesIO`
        """
        outfile_name = helper.save(model)

        a = BytesIO()
        a.seek(0, 0)
        with open(outfile_name, 'rb') as f:
            a.write(f.read())
        os.unlink(outfile_name)
        return a

    def get_model(self, id):
        """ Download model with id 'id' from server.

        :param id: The model id.
        :type id: str
        :return: A BytesIO object containing the model.
        :rtype: :class:`io.BytesIO`, None if model does not exist.
        """

        data = BytesIO()
        data.seek(0, 0)

        parts = self.Download(fedn.ModelRequest(id=id), self)
        for part in parts:
            if part.status == fedn.ModelStatus.IN_PROGRESS:
                data.write(part.data)

            if part.status == fedn.ModelStatus.OK:
                return data
            if part.status == fedn.ModelStatus.FAILED:
                return None

    def set_model(self, model, id):
        """ Upload model to server.

        :param model: A model object (BytesIO)
        :type model: :class:`io.BytesIO`
        :param id: The model id.
        :type id: str
        """
        if not isinstance(model, BytesIO):
            bt = BytesIO()

            written_total = 0
            for d in model.stream(32 * 1024):
                written = bt.write(d)
                written_total += written
        else:
            bt = model

        bt.seek(0, 0)

        def upload_request_generator(mdl):
            """

            :param mdl:
            """
            while True:
                b = mdl.read(CHUNK_SIZE)
                if b:
                    result = fedn.ModelRequest(
                        data=b, id=id, status=fedn.ModelStatus.IN_PROGRESS)
                else:
                    result = fedn.ModelRequest(
                        id=id, data=None, status=fedn.ModelStatus.OK)
                yield result
                if not b:
                    break

        # TODO: Check result
        _ = self.Upload(upload_request_generator(bt), self)

    # Model Service
    def Upload(self, request_iterator, context):
        """ RPC endpoints for uploading a model.

        :param request_iterator: The model request iterator.
        :type request_iterator: :class:`fedn.common.net.grpc.fedn_pb2.ModelRequest`
        :param context: The context object (unused)
        :type context: :class:`grpc._server._Context`
        :return: A model response object.
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.ModelResponse`
        """

        result = None
        for request in request_iterator:
            if request.status == fedn.ModelStatus.IN_PROGRESS:
                self.models.get_ptr(request.id).write(request.data)
                self.models.set_meta(request.id, fedn.ModelStatus.IN_PROGRESS)

            if request.status == fedn.ModelStatus.OK and not request.data:
                result = fedn.ModelResponse(id=request.id, status=fedn.ModelStatus.OK,
                                            message="Got model successfully.")
                # self.models_metadata.update({request.id: fedn.ModelStatus.OK})
                self.models.set_meta(request.id, fedn.ModelStatus.OK)
                self.models.get_ptr(request.id).flush()
                self.models.get_ptr(request.id).close()
                return result

    def Download(self, request, context):
        """ RPC endpoints for downloading a model.

        :param request: The model request object.
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.ModelRequest`
        :param context: The context object (unused)
        :type context: :class:`grpc._server._Context`
        :return: A model response iterator.
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.ModelResponse`
        """
        try:
            if self.models.get_meta(request.id) != fedn.ModelStatus.OK:
                print("Error file is not ready", flush=True)
                yield fedn.ModelResponse(id=request.id, data=None, status=fedn.ModelStatus.FAILED)
        except Exception:
            print("Error file does not exist: {}".format(request.id), flush=True)
            yield fedn.ModelResponse(id=request.id, data=None, status=fedn.ModelStatus.FAILED)

        try:
            obj = self.models.get(request.id)
            with obj as f:
                while True:
                    piece = f.read(CHUNK_SIZE)
                    if len(piece) == 0:
                        yield fedn.ModelResponse(id=request.id, data=None, status=fedn.ModelStatus.OK)
                        return
                    yield fedn.ModelResponse(id=request.id, data=piece, status=fedn.ModelStatus.IN_PROGRESS)
        except Exception as e:
            print("Downloading went wrong: {} {}".format(request.id, e), flush=True)



import logging
import os


class Logger:
    """ Logger class for Fedn.

    :param log_level: The log level.
    :type log_level: int
    :param to_file: The name of the file to log to.
    :type to_file: str
    :param file_path: The path to the log file.
    :type file_path: str
    """

    def __init__(self, log_level=logging.DEBUG, to_file='', file_path=os.getcwd()):
        """ Initialize the logger."""
        root = logging.getLogger()
        root.setLevel(log_level)

        # sh = logging.StreamHandler(sys.stdout)
        sh = logging.StreamHandler()
        sh.setLevel(log_level)
        log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        formatter = logging.Formatter(log_format)
        sh.setFormatter(formatter)
        root.addHandler(sh)

        if to_file != '':
            fh = logging.FileHandler(os.path.join(
                file_path, '{}'.format(to_file)))
            fh.setFormatter(logging.Formatter(log_format))
            root.addHandler(fh)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: fedn/common/net/grpc/fedn.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1f\x66\x65\x64n/common/net/grpc/fedn.proto\x12\x04grpc\":\n\x08Response\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x10\n\x08response\x18\x02 \x01(\t\"\x8c\x02\n\x06Status\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x0e\n\x06status\x18\x02 \x01(\t\x12(\n\tlog_level\x18\x03 \x01(\x0e\x32\x15.grpc.Status.LogLevel\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\t\x12\x1e\n\x04type\x18\x07 \x01(\x0e\x32\x10.grpc.StatusType\x12\r\n\x05\x65xtra\x18\x08 \x01(\t\"B\n\x08LogLevel\x12\x08\n\x04INFO\x10\x00\x12\t\n\x05\x44\x45\x42UG\x10\x01\x12\x0b\n\x07WARNING\x10\x02\x12\t\n\x05\x45RROR\x10\x03\x12\t\n\x05\x41UDIT\x10\x04\"\xab\x01\n\x12ModelUpdateRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.grpc.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\t\x12\x0c\n\x04meta\x18\x07 \x01(\t\"\xaf\x01\n\x0bModelUpdate\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.grpc.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\x12\x17\n\x0fmodel_update_id\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\t\x12\x0c\n\x04meta\x18\x07 \x01(\t\"\xc5\x01\n\x16ModelValidationRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.grpc.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\t\x12\x0c\n\x04meta\x18\x07 \x01(\t\x12\x14\n\x0cis_inference\x18\x08 \x01(\x08\"\xa8\x01\n\x0fModelValidation\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.grpc.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\t\x12\x0c\n\x04meta\x18\x07 \x01(\t\"\x89\x01\n\x0cModelRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.grpc.Client\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\n\n\x02id\x18\x04 \x01(\t\x12!\n\x06status\x18\x05 \x01(\x0e\x32\x11.grpc.ModelStatus\"]\n\rModelResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\n\n\x02id\x18\x02 \x01(\t\x12!\n\x06status\x18\x03 \x01(\x0e\x32\x11.grpc.ModelStatus\x12\x0f\n\x07message\x18\x04 \x01(\t\"U\n\x15GetGlobalModelRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.grpc.Client\"h\n\x16GetGlobalModelResponse\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.grpc.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\")\n\tHeartbeat\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\"W\n\x16\x43lientAvailableMessage\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\t\"R\n\x12ListClientsRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x07\x63hannel\x18\x02 \x01(\x0e\x32\r.grpc.Channel\"*\n\nClientList\x12\x1c\n\x06\x63lient\x18\x01 \x03(\x0b\x32\x0c.grpc.Client\"0\n\x06\x43lient\x12\x18\n\x04role\x18\x01 \x01(\x0e\x32\n.grpc.Role\x12\x0c\n\x04name\x18\x02 \x01(\t\"m\n\x0fReassignRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.grpc.Client\x12\x0e\n\x06server\x18\x03 \x01(\t\x12\x0c\n\x04port\x18\x04 \x01(\r\"c\n\x10ReconnectRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.grpc.Client\x12\x11\n\treconnect\x18\x03 \x01(\r\"\'\n\tParameter\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"T\n\x0e\x43ontrolRequest\x12\x1e\n\x07\x63ommand\x18\x01 \x01(\x0e\x32\r.grpc.Command\x12\"\n\tparameter\x18\x02 \x03(\x0b\x32\x0f.grpc.Parameter\"F\n\x0f\x43ontrolResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\"\n\tparameter\x18\x02 \x03(\x0b\x32\x0f.grpc.Parameter\"R\n\x0eReportResponse\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.grpc.Client\x12\"\n\tparameter\x18\x02 \x03(\x0b\x32\x0f.grpc.Parameter\"\x13\n\x11\x43onnectionRequest\"<\n\x12\x43onnectionResponse\x12&\n\x06status\x18\x01 \x01(\x0e\x32\x16.grpc.ConnectionStatus*\x84\x01\n\nStatusType\x12\x07\n\x03LOG\x10\x00\x12\x18\n\x14MODEL_UPDATE_REQUEST\x10\x01\x12\x10\n\x0cMODEL_UPDATE\x10\x02\x12\x1c\n\x18MODEL_VALIDATION_REQUEST\x10\x03\x12\x14\n\x10MODEL_VALIDATION\x10\x04\x12\r\n\tINFERENCE\x10\x05*\x86\x01\n\x07\x43hannel\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x19\n\x15MODEL_UPDATE_REQUESTS\x10\x01\x12\x11\n\rMODEL_UPDATES\x10\x02\x12\x1d\n\x19MODEL_VALIDATION_REQUESTS\x10\x03\x12\x15\n\x11MODEL_VALIDATIONS\x10\x04\x12\n\n\x06STATUS\x10\x05*F\n\x0bModelStatus\x12\x06\n\x02OK\x10\x00\x12\x0f\n\x0bIN_PROGRESS\x10\x01\x12\x12\n\x0eIN_PROGRESS_OK\x10\x02\x12\n\n\x06\x46\x41ILED\x10\x03*8\n\x04Role\x12\n\n\x06WORKER\x10\x00\x12\x0c\n\x08\x43OMBINER\x10\x01\x12\x0b\n\x07REDUCER\x10\x02\x12\t\n\x05OTHER\x10\x03*J\n\x07\x43ommand\x12\x08\n\x04IDLE\x10\x00\x12\t\n\x05START\x10\x01\x12\t\n\x05PAUSE\x10\x02\x12\x08\n\x04STOP\x10\x03\x12\t\n\x05RESET\x10\x04\x12\n\n\x06REPORT\x10\x05*I\n\x10\x43onnectionStatus\x12\x11\n\rNOT_ACCEPTING\x10\x00\x12\r\n\tACCEPTING\x10\x01\x12\x13\n\x0fTRY_AGAIN_LATER\x10\x02\x32z\n\x0cModelService\x12\x33\n\x06Upload\x12\x12.grpc.ModelRequest\x1a\x13.grpc.ModelResponse(\x01\x12\x35\n\x08\x44ownload\x12\x12.grpc.ModelRequest\x1a\x13.grpc.ModelResponse0\x01\x32\xa9\x02\n\x07\x43ontrol\x12\x34\n\x05Start\x12\x14.grpc.ControlRequest\x1a\x15.grpc.ControlResponse\x12\x33\n\x04Stop\x12\x14.grpc.ControlRequest\x1a\x15.grpc.ControlResponse\x12\x37\n\tConfigure\x12\x14.grpc.ControlRequest\x1a\x14.grpc.ReportResponse\x12\x44\n\x15\x46lushAggregationQueue\x12\x14.grpc.ControlRequest\x1a\x15.grpc.ControlResponse\x12\x34\n\x06Report\x12\x14.grpc.ControlRequest\x1a\x14.grpc.ReportResponse2V\n\x07Reducer\x12K\n\x0eGetGlobalModel\x12\x1b.grpc.GetGlobalModelRequest\x1a\x1c.grpc.GetGlobalModelResponse2\xab\x03\n\tConnector\x12\x44\n\x14\x41llianceStatusStream\x12\x1c.grpc.ClientAvailableMessage\x1a\x0c.grpc.Status0\x01\x12*\n\nSendStatus\x12\x0c.grpc.Status\x1a\x0e.grpc.Response\x12?\n\x11ListActiveClients\x12\x18.grpc.ListClientsRequest\x1a\x10.grpc.ClientList\x12\x45\n\x10\x41\x63\x63\x65ptingClients\x12\x17.grpc.ConnectionRequest\x1a\x18.grpc.ConnectionResponse\x12\x30\n\rSendHeartbeat\x12\x0f.grpc.Heartbeat\x1a\x0e.grpc.Response\x12\x37\n\x0eReassignClient\x12\x15.grpc.ReassignRequest\x1a\x0e.grpc.Response\x12\x39\n\x0fReconnectClient\x12\x16.grpc.ReconnectRequest\x1a\x0e.grpc.Response2\xda\x04\n\x08\x43ombiner\x12T\n\x18ModelUpdateRequestStream\x12\x1c.grpc.ClientAvailableMessage\x1a\x18.grpc.ModelUpdateRequest0\x01\x12\x46\n\x11ModelUpdateStream\x12\x1c.grpc.ClientAvailableMessage\x1a\x11.grpc.ModelUpdate0\x01\x12\\\n\x1cModelValidationRequestStream\x12\x1c.grpc.ClientAvailableMessage\x1a\x1c.grpc.ModelValidationRequest0\x01\x12N\n\x15ModelValidationStream\x12\x1c.grpc.ClientAvailableMessage\x1a\x15.grpc.ModelValidation0\x01\x12\x42\n\x16SendModelUpdateRequest\x12\x18.grpc.ModelUpdateRequest\x1a\x0e.grpc.Response\x12\x34\n\x0fSendModelUpdate\x12\x11.grpc.ModelUpdate\x1a\x0e.grpc.Response\x12J\n\x1aSendModelValidationRequest\x12\x1c.grpc.ModelValidationRequest\x1a\x0e.grpc.Response\x12<\n\x13SendModelValidation\x12\x15.grpc.ModelValidation\x1a\x0e.grpc.Responseb\x06proto3')

_STATUSTYPE = DESCRIPTOR.enum_types_by_name['StatusType']
StatusType = enum_type_wrapper.EnumTypeWrapper(_STATUSTYPE)
_CHANNEL = DESCRIPTOR.enum_types_by_name['Channel']
Channel = enum_type_wrapper.EnumTypeWrapper(_CHANNEL)
_MODELSTATUS = DESCRIPTOR.enum_types_by_name['ModelStatus']
ModelStatus = enum_type_wrapper.EnumTypeWrapper(_MODELSTATUS)
_ROLE = DESCRIPTOR.enum_types_by_name['Role']
Role = enum_type_wrapper.EnumTypeWrapper(_ROLE)
_COMMAND = DESCRIPTOR.enum_types_by_name['Command']
Command = enum_type_wrapper.EnumTypeWrapper(_COMMAND)
_CONNECTIONSTATUS = DESCRIPTOR.enum_types_by_name['ConnectionStatus']
ConnectionStatus = enum_type_wrapper.EnumTypeWrapper(_CONNECTIONSTATUS)
LOG = 0
MODEL_UPDATE_REQUEST = 1
MODEL_UPDATE = 2
MODEL_VALIDATION_REQUEST = 3
MODEL_VALIDATION = 4
INFERENCE = 5
DEFAULT = 0
MODEL_UPDATE_REQUESTS = 1
MODEL_UPDATES = 2
MODEL_VALIDATION_REQUESTS = 3
MODEL_VALIDATIONS = 4
STATUS = 5
OK = 0
IN_PROGRESS = 1
IN_PROGRESS_OK = 2
FAILED = 3
WORKER = 0
COMBINER = 1
REDUCER = 2
OTHER = 3
IDLE = 0
START = 1
PAUSE = 2
STOP = 3
RESET = 4
REPORT = 5
NOT_ACCEPTING = 0
ACCEPTING = 1
TRY_AGAIN_LATER = 2


_RESPONSE = DESCRIPTOR.message_types_by_name['Response']
_STATUS = DESCRIPTOR.message_types_by_name['Status']
_MODELUPDATEREQUEST = DESCRIPTOR.message_types_by_name['ModelUpdateRequest']
_MODELUPDATE = DESCRIPTOR.message_types_by_name['ModelUpdate']
_MODELVALIDATIONREQUEST = DESCRIPTOR.message_types_by_name['ModelValidationRequest']
_MODELVALIDATION = DESCRIPTOR.message_types_by_name['ModelValidation']
_MODELREQUEST = DESCRIPTOR.message_types_by_name['ModelRequest']
_MODELRESPONSE = DESCRIPTOR.message_types_by_name['ModelResponse']
_GETGLOBALMODELREQUEST = DESCRIPTOR.message_types_by_name['GetGlobalModelRequest']
_GETGLOBALMODELRESPONSE = DESCRIPTOR.message_types_by_name['GetGlobalModelResponse']
_HEARTBEAT = DESCRIPTOR.message_types_by_name['Heartbeat']
_CLIENTAVAILABLEMESSAGE = DESCRIPTOR.message_types_by_name['ClientAvailableMessage']
_LISTCLIENTSREQUEST = DESCRIPTOR.message_types_by_name['ListClientsRequest']
_CLIENTLIST = DESCRIPTOR.message_types_by_name['ClientList']
_CLIENT = DESCRIPTOR.message_types_by_name['Client']
_REASSIGNREQUEST = DESCRIPTOR.message_types_by_name['ReassignRequest']
_RECONNECTREQUEST = DESCRIPTOR.message_types_by_name['ReconnectRequest']
_PARAMETER = DESCRIPTOR.message_types_by_name['Parameter']
_CONTROLREQUEST = DESCRIPTOR.message_types_by_name['ControlRequest']
_CONTROLRESPONSE = DESCRIPTOR.message_types_by_name['ControlResponse']
_REPORTRESPONSE = DESCRIPTOR.message_types_by_name['ReportResponse']
_CONNECTIONREQUEST = DESCRIPTOR.message_types_by_name['ConnectionRequest']
_CONNECTIONRESPONSE = DESCRIPTOR.message_types_by_name['ConnectionResponse']
_STATUS_LOGLEVEL = _STATUS.enum_types_by_name['LogLevel']
Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
  'DESCRIPTOR' : _RESPONSE,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.Response)
  })
_sym_db.RegisterMessage(Response)

Status = _reflection.GeneratedProtocolMessageType('Status', (_message.Message,), {
  'DESCRIPTOR' : _STATUS,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.Status)
  })
_sym_db.RegisterMessage(Status)

ModelUpdateRequest = _reflection.GeneratedProtocolMessageType('ModelUpdateRequest', (_message.Message,), {
  'DESCRIPTOR' : _MODELUPDATEREQUEST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ModelUpdateRequest)
  })
_sym_db.RegisterMessage(ModelUpdateRequest)

ModelUpdate = _reflection.GeneratedProtocolMessageType('ModelUpdate', (_message.Message,), {
  'DESCRIPTOR' : _MODELUPDATE,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ModelUpdate)
  })
_sym_db.RegisterMessage(ModelUpdate)

ModelValidationRequest = _reflection.GeneratedProtocolMessageType('ModelValidationRequest', (_message.Message,), {
  'DESCRIPTOR' : _MODELVALIDATIONREQUEST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ModelValidationRequest)
  })
_sym_db.RegisterMessage(ModelValidationRequest)

ModelValidation = _reflection.GeneratedProtocolMessageType('ModelValidation', (_message.Message,), {
  'DESCRIPTOR' : _MODELVALIDATION,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ModelValidation)
  })
_sym_db.RegisterMessage(ModelValidation)

ModelRequest = _reflection.GeneratedProtocolMessageType('ModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _MODELREQUEST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ModelRequest)
  })
_sym_db.RegisterMessage(ModelRequest)

ModelResponse = _reflection.GeneratedProtocolMessageType('ModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _MODELRESPONSE,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ModelResponse)
  })
_sym_db.RegisterMessage(ModelResponse)

GetGlobalModelRequest = _reflection.GeneratedProtocolMessageType('GetGlobalModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETGLOBALMODELREQUEST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.GetGlobalModelRequest)
  })
_sym_db.RegisterMessage(GetGlobalModelRequest)

GetGlobalModelResponse = _reflection.GeneratedProtocolMessageType('GetGlobalModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETGLOBALMODELRESPONSE,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.GetGlobalModelResponse)
  })
_sym_db.RegisterMessage(GetGlobalModelResponse)

Heartbeat = _reflection.GeneratedProtocolMessageType('Heartbeat', (_message.Message,), {
  'DESCRIPTOR' : _HEARTBEAT,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.Heartbeat)
  })
_sym_db.RegisterMessage(Heartbeat)

ClientAvailableMessage = _reflection.GeneratedProtocolMessageType('ClientAvailableMessage', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTAVAILABLEMESSAGE,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ClientAvailableMessage)
  })
_sym_db.RegisterMessage(ClientAvailableMessage)

ListClientsRequest = _reflection.GeneratedProtocolMessageType('ListClientsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTCLIENTSREQUEST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ListClientsRequest)
  })
_sym_db.RegisterMessage(ListClientsRequest)

ClientList = _reflection.GeneratedProtocolMessageType('ClientList', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTLIST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ClientList)
  })
_sym_db.RegisterMessage(ClientList)

Client = _reflection.GeneratedProtocolMessageType('Client', (_message.Message,), {
  'DESCRIPTOR' : _CLIENT,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.Client)
  })
_sym_db.RegisterMessage(Client)

ReassignRequest = _reflection.GeneratedProtocolMessageType('ReassignRequest', (_message.Message,), {
  'DESCRIPTOR' : _REASSIGNREQUEST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ReassignRequest)
  })
_sym_db.RegisterMessage(ReassignRequest)

ReconnectRequest = _reflection.GeneratedProtocolMessageType('ReconnectRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECONNECTREQUEST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ReconnectRequest)
  })
_sym_db.RegisterMessage(ReconnectRequest)

Parameter = _reflection.GeneratedProtocolMessageType('Parameter', (_message.Message,), {
  'DESCRIPTOR' : _PARAMETER,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.Parameter)
  })
_sym_db.RegisterMessage(Parameter)

ControlRequest = _reflection.GeneratedProtocolMessageType('ControlRequest', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLREQUEST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ControlRequest)
  })
_sym_db.RegisterMessage(ControlRequest)

ControlResponse = _reflection.GeneratedProtocolMessageType('ControlResponse', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLRESPONSE,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ControlResponse)
  })
_sym_db.RegisterMessage(ControlResponse)

ReportResponse = _reflection.GeneratedProtocolMessageType('ReportResponse', (_message.Message,), {
  'DESCRIPTOR' : _REPORTRESPONSE,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ReportResponse)
  })
_sym_db.RegisterMessage(ReportResponse)

ConnectionRequest = _reflection.GeneratedProtocolMessageType('ConnectionRequest', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTIONREQUEST,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ConnectionRequest)
  })
_sym_db.RegisterMessage(ConnectionRequest)

ConnectionResponse = _reflection.GeneratedProtocolMessageType('ConnectionResponse', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTIONRESPONSE,
  '__module__' : 'fedn.common.net.grpc.fedn_pb2'
  # @@protoc_insertion_point(class_scope:grpc.ConnectionResponse)
  })
_sym_db.RegisterMessage(ConnectionResponse)

_MODELSERVICE = DESCRIPTOR.services_by_name['ModelService']
_CONTROL = DESCRIPTOR.services_by_name['Control']
_REDUCER = DESCRIPTOR.services_by_name['Reducer']
_CONNECTOR = DESCRIPTOR.services_by_name['Connector']
_COMBINER = DESCRIPTOR.services_by_name['Combiner']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATUSTYPE._serialized_start=2412
  _STATUSTYPE._serialized_end=2544
  _CHANNEL._serialized_start=2547
  _CHANNEL._serialized_end=2681
  _MODELSTATUS._serialized_start=2683
  _MODELSTATUS._serialized_end=2753
  _ROLE._serialized_start=2755
  _ROLE._serialized_end=2811
  _COMMAND._serialized_start=2813
  _COMMAND._serialized_end=2887
  _CONNECTIONSTATUS._serialized_start=2889
  _CONNECTIONSTATUS._serialized_end=2962
  _RESPONSE._serialized_start=41
  _RESPONSE._serialized_end=99
  _STATUS._serialized_start=102
  _STATUS._serialized_end=370
  _STATUS_LOGLEVEL._serialized_start=304
  _STATUS_LOGLEVEL._serialized_end=370
  _MODELUPDATEREQUEST._serialized_start=373
  _MODELUPDATEREQUEST._serialized_end=544
  _MODELUPDATE._serialized_start=547
  _MODELUPDATE._serialized_end=722
  _MODELVALIDATIONREQUEST._serialized_start=725
  _MODELVALIDATIONREQUEST._serialized_end=922
  _MODELVALIDATION._serialized_start=925
  _MODELVALIDATION._serialized_end=1093
  _MODELREQUEST._serialized_start=1096
  _MODELREQUEST._serialized_end=1233
  _MODELRESPONSE._serialized_start=1235
  _MODELRESPONSE._serialized_end=1328
  _GETGLOBALMODELREQUEST._serialized_start=1330
  _GETGLOBALMODELREQUEST._serialized_end=1415
  _GETGLOBALMODELRESPONSE._serialized_start=1417
  _GETGLOBALMODELRESPONSE._serialized_end=1521
  _HEARTBEAT._serialized_start=1523
  _HEARTBEAT._serialized_end=1564
  _CLIENTAVAILABLEMESSAGE._serialized_start=1566
  _CLIENTAVAILABLEMESSAGE._serialized_end=1653
  _LISTCLIENTSREQUEST._serialized_start=1655
  _LISTCLIENTSREQUEST._serialized_end=1737
  _CLIENTLIST._serialized_start=1739
  _CLIENTLIST._serialized_end=1781
  _CLIENT._serialized_start=1783
  _CLIENT._serialized_end=1831
  _REASSIGNREQUEST._serialized_start=1833
  _REASSIGNREQUEST._serialized_end=1942
  _RECONNECTREQUEST._serialized_start=1944
  _RECONNECTREQUEST._serialized_end=2043
  _PARAMETER._serialized_start=2045
  _PARAMETER._serialized_end=2084
  _CONTROLREQUEST._serialized_start=2086
  _CONTROLREQUEST._serialized_end=2170
  _CONTROLRESPONSE._serialized_start=2172
  _CONTROLRESPONSE._serialized_end=2242
  _REPORTRESPONSE._serialized_start=2244
  _REPORTRESPONSE._serialized_end=2326
  _CONNECTIONREQUEST._serialized_start=2328
  _CONNECTIONREQUEST._serialized_end=2347
  _CONNECTIONRESPONSE._serialized_start=2349
  _CONNECTIONRESPONSE._serialized_end=2409
  _MODELSERVICE._serialized_start=2964
  _MODELSERVICE._serialized_end=3086
  _CONTROL._serialized_start=3089
  _CONTROL._serialized_end=3386
  _REDUCER._serialized_start=3388
  _REDUCER._serialized_end=3474
  _CONNECTOR._serialized_start=3477
  _CONNECTOR._serialized_end=3904
  _COMBINER._serialized_start=3907
  _COMBINER._serialized_end=4509
# @@protoc_insertion_point(module_scope)



import tempfile

import numpy as np

from .helperbase import HelperBase


class Helper(HelperBase):
    """ FEDn helper class for numpy arrays. """

    def increment_average(self, model, model_next, n):
        """ Update an incremental average.

        :param model: Current model weights.
        :type model: numpy array.
        :param model_next: New model weights.
        :type model_next: numpy array.
        :param n: Number of examples in new model.
        :type n: int
        :return: Incremental weighted average of model weights.
        :rtype: :class:`numpy.array`
        """
        return np.add(model, (model_next - model) / n)

    def save(self, model, path=None):
        """Serialize weights/parameters to file.

        :param model: Weights/parameters in numpy array format.
        :type model: numpy array.
        :param path: Path to file.
        :type path: str
        :return: Path to file.
        :rtype: str
        """
        if not path:
            _, path = tempfile.mkstemp()
        np.savetxt(path, model)
        return path

    def load(self, path):
        """Load weights/parameters from file or filelike.

        :param path: Path to file.
        :type path: str
        :return: Weights/parameters in numpy array format.
        :rtype: :class:`numpy.array`
        """
        model = np.loadtxt(path)
        return model



import os
from io import BytesIO

import fedn.common.net.grpc.fedn_pb2 as fedn
from fedn.common.storage.models.modelstorage import ModelStorage

CHUNK_SIZE = 1024 * 1024


class TempModelStorage(ModelStorage):
    """

    """

    def __init__(self):

        self.default_dir = os.environ.get(
            'FEDN_MODEL_DIR', '/tmp/models')  # set default to tmp
        if not os.path.exists(self.default_dir):
            os.makedirs(self.default_dir)

        # TODO should read in already existing temp models if crashed? or fetch new on demand (default)

        # self.models = defaultdict(io.BytesIO)
        self.models = {}
        self.models_metadata = {}

    def exist(self, model_id):
        """

        :param model_id:
        :return:
        """
        if model_id in self.models.keys():
            return True
        return False

    def get(self, model_id):
        """

        :param model_id:
        :return:
        """
        try:
            if self.models_metadata[model_id] != fedn.ModelStatus.OK:
                print("File not ready! Try again", flush=True)
                return None
        except KeyError:
            print("No such model have been made available yet!", flush=True)
            return None

        obj = BytesIO()
        with open(os.path.join(self.default_dir, str(model_id)), 'rb') as f:
            obj.write(f.read())

        obj.seek(0, 0)
        return obj

    def get_ptr(self, model_id):
        """

        :param model_id:
        :return:
        """
        try:
            f = self.models[model_id]['file']
        except KeyError:
            f = open(os.path.join(self.default_dir, str(model_id)), "wb")

        self.models[model_id] = {'file': f}
        return self.models[model_id]['file']

    def get_meta(self, model_id):
        """

        :param model_id:
        :return:
        """
        return self.models_metadata[model_id]

    def set_meta(self, model_id, model_metadata):
        """

        :param model_id:
        :param model_metadata:
        """
        self.models_metadata.update({model_id: model_metadata})

    # Delete model from disk
    def delete(self, model_id):
        """ Delete model from temp disk/storage

        :param model_id: model id
        :type model_id: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        try:
            os.remove(os.path.join(self.default_dir, str(model_id)))
            print("TEMPMODELSTORAGE: Deleted model with id: {}".format(model_id), flush=True)
            # Delete id from metadata and models dict
            del self.models_metadata[model_id]
            del self.models[model_id]
        except FileNotFoundError:
            print("Could not delete model from disk. File not found!", flush=True)
            return False
        return True

    # Delete all models from disk
    def delete_all(self):
        """ Delete all models from temp disk/storage

        :return: True if successful, False otherwise
        :rtype: bool
        """
        ids_pop = []
        for model_id in self.models.keys():
            try:
                os.remove(os.path.join(self.default_dir, str(model_id)))
                print("TEMPMODELSTORAGE: Deleted model with id: {}".format(model_id), flush=True)
                # Add id to list of ids to pop/delete from metadata and models dict
                ids_pop.append(model_id)
            except FileNotFoundError:
                print("TEMPMODELSTORAGE: Could not delete model {} from disk. File not found!".format(model_id), flush=True)
        # Remove id from metadata and models dict
        for model_id in ids_pop:
            del self.models_metadata[model_id]
            del self.models[model_id]
        return True



import time
import uuid

import click
import yaml

from fedn.common.exceptions import InvalidClientConfig
from fedn.network.clients.client import Client
from fedn.network.combiner.server import Combiner
from fedn.network.dashboard.restservice import (decode_auth_token,
                                                encode_auth_token)
from fedn.network.reducer import Reducer
from fedn.network.statestore.mongostatestore import MongoStateStore

from .main import main


def get_statestore_config_from_file(init):
    """

    :param init:
    :return:
    """
    with open(init, 'r') as file:
        try:
            settings = dict(yaml.safe_load(file))
            return settings
        except yaml.YAMLError as e:
            raise (e)


def check_helper_config_file(config):
    control = config['control']
    try:
        helper = control["helper"]
    except KeyError:
        print("--local-package was used, but no helper was found in --init settings file.", flush=True)
        exit(-1)
    return helper


def apply_config(config):
    """Parse client config from file.

    Override configs from the CLI with settings in config file.

    :param config: Client config (dict).
    """
    with open(config['init'], 'r') as file:
        try:
            settings = dict(yaml.safe_load(file))
        except Exception:
            print('Failed to read config from settings file, exiting.', flush=True)
            return

    for key, val in settings.items():
        config[key] = val


def validate_client_config(config):
    """Validate client configuration.

    :param config: Client config (dict).
    """

    try:
        if config['discover_host'] is None or \
                config['discover_host'] == '':
            raise InvalidClientConfig("Missing required configuration: discover_host")
        if 'discover_port' not in config.keys():
            config['discover_port'] = None
    except Exception:
        raise InvalidClientConfig("Could not load config from file. Check config")


@main.group('run')
@click.pass_context
def run_cmd(ctx):
    """

    :param ctx:
    """
    # if daemon:
    #    print('{} NYI should run as daemon...'.format(__file__))
    pass


@run_cmd.command('client')
@click.option('-d', '--discoverhost', required=False, help='Hostname for discovery services(reducer).')
@click.option('-p', '--discoverport', required=False, help='Port for discovery services (reducer).')
@click.option('--token', required=False, help='Set token provided by reducer if enabled')
@click.option('-n', '--name', required=False, default="client" + str(uuid.uuid4())[:8])
@click.option('-i', '--client_id', required=False)
@click.option('--local-package', is_flag=True, help='Enable local compute package')
@click.option('--force-ssl', is_flag=True, help='Force SSL/TLS for REST service')
@click.option('-u', '--dry-run', required=False, default=False)
@click.option('-s', '--secure', required=False, default=False)
@click.option('-pc', '--preshared-cert', required=False, default=False)
@click.option('-v', '--verify', is_flag=True, help='Verify SSL/TLS for REST service')
@click.option('-c', '--preferred-combiner', required=False, default=False)
@click.option('-va', '--validator', required=False, default=True)
@click.option('-tr', '--trainer', required=False, default=True)
@click.option('-in', '--init', required=False, default=None,
              help='Set to a filename to (re)init client from file state.')
@click.option('-l', '--logfile', required=False, default='{}-client.log'.format(time.strftime("%Y%m%d-%H%M%S")),
              help='Set logfile for client log to file.')
@click.option('--heartbeat-interval', required=False, default=2)
@click.option('--reconnect-after-missed-heartbeat', required=False, default=30)
@click.pass_context
def client_cmd(ctx, discoverhost, discoverport, token, name, client_id, local_package, force_ssl, dry_run, secure, preshared_cert,
               verify, preferred_combiner, validator, trainer, init, logfile, heartbeat_interval, reconnect_after_missed_heartbeat):
    """

    :param ctx:
    :param discoverhost:
    :param discoverport:
    :param token:
    :param name:
    :param client_id:
    :param remote:
    :param dry_run:
    :param secure:
    :param preshared_cert:
    :param verify_cert:
    :param preferred_combiner:
    :param init:
    :param logfile:
    :param hearbeat_interval
    :param reconnect_after_missed_heartbeat
    :return:
    """
    remote = False if local_package else True
    config = {'discover_host': discoverhost, 'discover_port': discoverport, 'token': token, 'name': name,
              'client_id': client_id, 'remote_compute_context': remote, 'force_ssl': force_ssl, 'dry_run': dry_run, 'secure': secure,
              'preshared_cert': preshared_cert, 'verify': verify, 'preferred_combiner': preferred_combiner,
              'validator': validator, 'trainer': trainer, 'init': init, 'logfile': logfile, 'heartbeat_interval': heartbeat_interval,
              'reconnect_after_missed_heartbeat': reconnect_after_missed_heartbeat}

    if init:
        apply_config(config)

    validate_client_config(config)

    client = Client(config)
    client.run()


@run_cmd.command('dashboard')
@click.option('-h', '--host', required=False)
@click.option('-p', '--port', required=False, default='8090', show_default=True)
@click.option('-k', '--secret-key', required=False, help='Set secret key to enable jwt token authentication.')
@click.option('-l', '--local-package', is_flag=True, help='Enable use of local compute package')
@click.option('-n', '--name', required=False, default="reducer" + str(uuid.uuid4())[:8], help='Set service name')
@click.option('-in', '--init', required=True, default=None,
              help='Set to a filename to (re)init reducer state from file.')
@click.pass_context
def dashboard_cmd(ctx, host, port, secret_key, local_package, name, init):
    """ Start the dashboard service.

    :param ctx: Click context.
    :param discoverhost: Hostname for discovery services (dashboard).
    :param discoverport: Port for discovery services (dashboard).
    :param secret_key: Set secret key to enable jwt token authentication.
    :param local_package: Enable use of local compute package.
    :param name: Set service name.
    :param init: Set to a filename to (re)init config state from file.
    """
    remote = False if local_package else True
    config = {'host': host, 'port': port, 'secret_key': secret_key,
              'name': name, 'remote_compute_package': remote, 'init': init}

    # Read settings from config file
    try:
        fedn_config = get_statestore_config_from_file(config['init'])
    except Exception as e:
        print('Failed to read config from settings file, exiting.', flush=True)
        print(e, flush=True)
        exit(-1)

    if not remote:
        _ = check_helper_config_file(fedn_config)

    try:
        network_id = fedn_config['network_id']
    except KeyError:
        print("No network_id in config, please specify the control network id.", flush=True)
        exit(-1)

    # Obtain state from database, in case already initialized (service restart)
    statestore_config = fedn_config['statestore']
    if statestore_config['type'] == 'MongoDB':
        statestore = MongoStateStore(
            network_id, statestore_config['mongo_config'], fedn_config['storage'])
    else:
        print("Unsupported statestore type, exiting. ", flush=True)
        exit(-1)

    # Enable JWT token authentication.
    if config['secret_key']:
        # If we already have a valid token in statestore config, use that one.
        existing_config = statestore.get_reducer()
        if existing_config:
            try:
                existing_config = statestore.get_reducer()
                current_token = existing_config['token']
                status = decode_auth_token(current_token, config['secret_key'])
                if status != 'Success':
                    token = encode_auth_token(config['secret_key'])
                    config['token'] = token
            except Exception:
                raise

        else:
            token = encode_auth_token(config['secret_key'])
            config['token'] = token
    try:
        statestore.set_reducer(config)
    except Exception:
        print("Failed to set reducer config in statestore, exiting.", flush=True)
        exit(-1)

    # Configure storage backend.
    try:
        statestore.set_storage_backend(fedn_config['storage'])
    except KeyError:
        print("storage configuration missing in statestore_config.", flush=True)
        exit(-1)
    except Exception:
        print("Failed to set storage config in statestore, exiting.", flush=True)
        exit(-1)

    reducer = Reducer(statestore)
    reducer.run()


@run_cmd.command('combiner')
@click.option('-d', '--discoverhost', required=False, help='Hostname for discovery services (reducer).')
@click.option('-p', '--discoverport', required=False, help='Port for discovery services (reducer).')
@click.option('-t', '--token', required=False, help='Set token provided by reducer if enabled')
@click.option('-n', '--name', required=False, default="combiner" + str(uuid.uuid4())[:8], help='Set name for combiner.')
@click.option('-h', '--host', required=False, default="combiner", help='Set hostname.')
@click.option('-i', '--port', required=False, default=12080, help='Set port.')
@click.option('-f', '--fqdn', required=False, default=None, help='Set fully qualified domain name')
@click.option('-s', '--secure', is_flag=True, help='Enable SSL/TLS encrypted gRPC channels.')
@click.option('-v', '--verify', is_flag=True, help='Verify SSL/TLS for REST discovery service (reducer)')
@click.option('-c', '--max_clients', required=False, default=30, help='The maximal number of client connections allowed.')
@click.option('-in', '--init', required=False, default=None,
              help='Path to configuration file to (re)init combiner.')
@click.option('-a', '--aggregator', required=False, default='fedavg', help='Filename of the aggregator module to use.')
@click.pass_context
def combiner_cmd(ctx, discoverhost, discoverport, token, name, host, port, fqdn, secure, verify, max_clients, init, aggregator):
    """

    :param ctx:
    :param discoverhost:
    :param discoverport:
    :param token:
    :param name:
    :param hostname:
    :param port:
    :param secure:
    :param max_clients:
    :param init:
    """
    config = {'discover_host': discoverhost, 'discover_port': discoverport, 'token': token, 'host': host,
              'port': port, 'fqdn': fqdn, 'name': name, 'secure': secure, 'verify': verify, 'max_clients': max_clients,
              'init': init, 'aggregator': aggregator}

    if config['init']:
        apply_config(config)

    combiner = Combiner(config)
    combiner.run()



import os
import re
import threading
import time
from datetime import datetime

from fedn.common.security.certificatemanager import CertificateManager
from fedn.network.controller.control import Control
from fedn.network.dashboard.restservice import ReducerRestService
from fedn.network.state import ReducerStateToString

VALID_NAME_REGEX = '^[a-zA-Z0-9_-]*$'


class InvalidReducerConfiguration(Exception):
    pass


class MissingReducerConfiguration(Exception):
    pass


class Reducer:
    """ A class used to instantiate the Reducer service.

    :param statestore: The backend statestore object.
    :type statestore: :class:`fedn.network.statestore.statestorebase.StateStoreBase`
    """

    def __init__(self, statestore):
        """ Constructor"""

        self.statestore = statestore
        config = self.statestore.get_reducer()
        if not config:
            print("REDUCER: Failed to retrive Reducer config, exiting.")
            raise MissingReducerConfiguration()

        # Validate reducer name
        match = re.search(VALID_NAME_REGEX, config['name'])
        if not match:
            raise ValueError('Unallowed character in reducer name. Allowed characters: a-z, A-Z, 0-9, _, -.')
        self.name = config['name']

        # The certificate manager is a utility that generates (self-signed) certificates.
        self.certificate_manager = CertificateManager(os.getcwd() + "/certs/")

        self.control = Control(self.statestore)

        self.rest = ReducerRestService(
            config, self.control, self.statestore, self.certificate_manager)

    def run(self):
        """Start REST service and control loop."""

        threading.Thread(target=self.control_loop, daemon=True).start()
        self.rest.run()

    def control_loop(self):
        """Manage and report the state of the Reducer."""

        try:
            old_state = self.control.state()

            t1 = datetime.now()
            while True:
                time.sleep(1)
                if old_state != self.control.state():
                    delta = datetime.now() - t1
                    print(
                        "Reducer in state {} for {} seconds. Entering {} state".format(ReducerStateToString(old_state),
                                                                                       delta.seconds,
                                                                                       ReducerStateToString(
                                                                                       self.control.state())),
                        flush=True)
                    t1 = datetime.now()
                    old_state = self.control.state()

        except (KeyboardInterrupt, SystemExit):
            print("Exiting..", flush=True)



import queue
import random
import sys
import time
import uuid

from fedn.network.combiner.aggregators.aggregatorbase import get_aggregator
from fedn.utils.helpers import get_helper


class ModelUpdateError(Exception):
    pass


class RoundController:
    """ Round controller.

    The round controller recieves round configurations from the global controller
    and coordinates model updates and aggregation, and model validations.

    :param aggregator_name: The name of the aggregator plugin module.
    :type aggregator_name: str
    :param storage: Model repository for :class: `fedn.network.combiner.Combiner`
    :type storage: class: `fedn.common.storage.s3.s3repo.S3ModelRepository`
    :param server: A handle to the Combiner class :class: `fedn.network.combiner.Combiner`
    :type server: class: `fedn.network.combiner.Combiner`
    :param modelservice: A handle to the model service :class: `fedn.network.combiner.modelservice.ModelService`
    :type modelservice: class: `fedn.network.combiner.modelservice.ModelService`
    """

    def __init__(self, aggregator_name, storage, server, modelservice):
        """ Initialize the RoundController."""

        self.round_configs = queue.Queue()
        self.storage = storage
        self.server = server
        self.modelservice = modelservice
        self.aggregator = get_aggregator(aggregator_name, self.storage, self.server, self.modelservice, self)

    def push_round_config(self, round_config):
        """Add a round_config (job description) to the inbox.

        :param round_config: A dict containing the round configuration (from global controller).
        :type round_config: dict
        :return: A job id (universally unique identifier) for the round.
        :rtype: str
        """
        try:
            round_config['_job_id'] = str(uuid.uuid4())
            self.round_configs.put(round_config)
        except Exception:
            self.server.report_status(
                "ROUNDCONTROL: Failed to push round config.", flush=True)
            raise
        return round_config['_job_id']

    def load_model_update(self, helper, model_id):
        """Load model update in its native format.

        :param helper: An instance of :class: `fedn.utils.helpers.HelperBase`, ML framework specific helper, defaults to None
        :type helper: class: `fedn.utils.helpers.HelperBase`
        :param model_id: The ID of the model update, UUID in str format
        :type model_id: str
        """

        model_str = self.load_model_update_str(model_id)
        if model_str:
            try:
                model = self.modelservice.load_model_from_BytesIO(model_str.getbuffer(), helper)
            except IOError:
                self.server.report_status(
                    "AGGREGATOR({}): Failed to load model!".format(self.name))
        else:
            raise ModelUpdateError("Failed to load model.")

        return model

    def load_model_update_str(self, model_id, retry=3):
        """Load model update object and return it as BytesIO.

        :param model_id: The ID of the model
        :type model_id: str
        :param retry: number of times retrying load model update, defaults to 3
        :type retry: int, optional
        :return: Updated model
        :rtype: class: `io.BytesIO`
        """
        # Try reading model update from local disk/combiner memory
        model_str = self.modelservice.models.get(model_id)
        # And if we cannot access that, try downloading from the server
        if model_str is None:
            model_str = self.modelservice.get_model(model_id)
            # TODO: use retrying library
            tries = 0
            while tries < retry:
                tries += 1
                if not model_str or sys.getsizeof(model_str) == 80:
                    self.server.report_status(
                        "ROUNDCONTROL: Model download failed. retrying", flush=True)

                    time.sleep(1)
                    model_str = self.modelservice.get_model(model_id)

        return model_str

    def waitforit(self, config, buffer_size=100, polling_interval=0.1):
        """ Defines the policy for how long the server should wait before starting to aggregate models.

        The policy is as follows:
            1. Wait a maximum of time_window time until the round times out.
            2. Terminate if a preset number of model updates (buffer_size) are in the queue.

        :param config: The round config object
        :type config: dict
        :param buffer_size: The number of model updates to wait for before starting aggregation, defaults to 100
        :type buffer_size: int, optional
        :param polling_interval: The polling interval, defaults to 0.1
        :type polling_interval: float, optional
        """

        time_window = float(config['round_timeout'])

        tt = 0.0
        while tt < time_window:
            if self.aggregator.model_updates.qsize() >= buffer_size:
                break

            time.sleep(polling_interval)
            tt += polling_interval

    def _training_round(self, config, clients):
        """Send model update requests to clients and aggregate results.

        :param config: The round config object (passed to the client).
        :type config: dict
        :param clients: clients to participate in the training round
        :type clients: list
        :return: an aggregated model and associated metadata
        :rtype: model, dict
        """

        self.server.report_status(
            "ROUNDCONTROL: Initiating training round, participating clients: {}".format(clients))

        meta = {}
        meta['nr_expected_updates'] = len(clients)
        meta['nr_required_updates'] = int(config['clients_required'])
        meta['timeout'] = float(config['round_timeout'])

        # Request model updates from all active clients.
        self.server.request_model_update(config, clients=clients)

        # If buffer_size is -1 (default), the round terminates when/if all clients have completed.
        if int(config['buffer_size']) == -1:
            buffer_size = len(clients)
        else:
            buffer_size = int(config['buffer_size'])

        # Wait / block until the round termination policy has been met.
        self.waitforit(config, buffer_size=buffer_size)

        tic = time.time()
        model = None
        data = None

        try:
            helper = get_helper(config['helper_type'])
            print("ROUNDCONTROL: Config delete_models_storage: {}".format(config['delete_models_storage']), flush=True)
            if config['delete_models_storage'] == 'True':
                delete_models = True
            else:
                delete_models = False
            model, data = self.aggregator.combine_models(helper=helper,
                                                         delete_models=delete_models)
        except Exception as e:
            print("AGGREGATION FAILED AT COMBINER! {}".format(e), flush=True)

        meta['time_combination'] = time.time() - tic
        meta['aggregation_time'] = data
        return model, meta

    def _validation_round(self, config, clients, model_id):
        """Send model validation requests to clients.

        :param config: The round config object (passed to the client).
        :type config: dict
        :param clients: clients to send validation requests to
        :type clients: list
        :param model_id: The ID of the model to validate
        :type model_id: str
        """
        self.server.request_model_validation(model_id, config, clients)

    def stage_model(self, model_id, timeout_retry=3, retry=2):
        """Download a model from persistent storage and set in modelservice.

        :param model_id: ID of the model update object to stage.
        :type model_id: str
        :param timeout_retry: Sleep before retrying download again(sec), defaults to 3
        :type timeout_retry: int, optional
        :param retry: Number of retries, defaults to 2
        :type retry: int, optional
        """

        # If the model is already in memory at the server we do not need to do anything.
        if self.modelservice.models.exist(model_id):
            print("ROUNDCONTROL: Model already exists in memory, skipping model staging.", flush=True)
            return
        print("ROUNDCONTROL: Model Staging, fetching model from storage...", flush=True)
        # If not, download it and stage it in memory at the combiner.
        tries = 0
        while True:
            try:
                model = self.storage.get_model_stream(model_id)
                if model:
                    break
            except Exception:
                self.server.report_status("ROUNDCONTROL: Could not fetch model from storage backend, retrying.",
                                          flush=True)
                time.sleep(timeout_retry)
                tries += 1
                if tries > retry:
                    self.server.report_status(
                        "ROUNDCONTROL: Failed to stage model {} from storage backend!".format(model_id), flush=True)
                    return

        self.modelservice.set_model(model, model_id)

    def _assign_round_clients(self, n, type="trainers"):
        """ Obtain a list of clients(trainers or validators) to ask for updates in this round.

        :param n: Size of a random set taken from active trainers(clients), if n > "active trainers" all is used
        :type n: int
        :param type: type of clients, either "trainers" or "validators", defaults to "trainers"
        :type type: str, optional
        :return: Set of clients
        :rtype: list
        """

        if type == "validators":
            clients = self.server.get_active_validators()
        elif type == "trainers":
            clients = self.server.get_active_trainers()
        else:
            self.server.report_status(
                "ROUNDCONTROL(ERROR): {} is not a supported type of client".format(type), flush=True)
            raise

        # If the number of requested trainers exceeds the number of available, use all available.
        if n > len(clients):
            n = len(clients)

        # If not, we pick a random subsample of all available clients.
        clients = random.sample(clients, n)

        return clients

    def _check_nr_round_clients(self, config, timeout=0.0):
        """Check that the minimal number of clients required to start a round are available.

        :param config: The round config object.
        :type config: dict
        :param timeout: Timeout in seconds, defaults to 0.0
        :type timeout: float, optional
        :return: True if the required number of clients are available, False otherwise.
        :rtype: bool
        """

        ready = False
        t = 0.0
        while not ready:
            active = self.server.nr_active_trainers()

            if active >= int(config['clients_requested']):
                return True
            else:
                self.server.report_status("waiting for {} clients to get started, currently: {}".format(
                    int(config['clients_requested']) - active,
                    active), flush=True)
            if t >= timeout:
                if active >= int(config['clients_required']):
                    return True
                else:
                    return False

            time.sleep(1.0)
            t += 1.0

        return ready

    def execute_validation_round(self, round_config):
        """ Coordinate validation rounds as specified in config.

        :param round_config: The round config object.
        :type round_config: dict
        """
        model_id = round_config['model_id']
        self.server.report_status(
            "COMBINER orchestrating validation of model {}".format(model_id))
        self.stage_model(model_id)
        validators = self._assign_round_clients(
            self.server.max_clients, type="validators")
        self._validation_round(round_config, validators, model_id)

    def execute_training_round(self, config):
        """ Coordinates clients to execute training tasks.

        :param config: The round config object.
        :type config: dict
        :return: metadata about the training round.
        :rtype: dict
        """

        self.server.report_status(
            "ROUNDCONTROL: Processing training round,  job_id {}".format(config['_job_id']), flush=True)

        data = {}
        data['config'] = config
        data['round_id'] = config['round_id']

        # Make sure the model to update is available on this combiner.
        self.stage_model(config['model_id'])

        clients = self._assign_round_clients(self.server.max_clients)
        model, meta = self._training_round(config, clients)
        data['data'] = meta

        if model is None:
            self.server.report_status(
                "\t Failed to update global model in round {0}!".format(config['round_id']))

        if model is not None:
            helper = get_helper(config['helper_type'])
            a = self.modelservice.serialize_model_to_BytesIO(model, helper)
            # Send aggregated model to server
            model_id = str(uuid.uuid4())
            self.modelservice.set_model(a, model_id)
            a.close()
            data['model_id'] = model_id

            self.server.report_status(
                "ROUNDCONTROL: TRAINING ROUND COMPLETED. Aggregated model id: {}, Job id: {}".format(model_id, config['_job_id']), flush=True)

        return data

    def run(self, polling_interval=1.0):
        """ Main control loop. Execute rounds based on round config on the queue.

        :param polling_interval: The polling interval in seconds for checking if a new job/config is available.
        :type polling_interval: float
        """
        try:
            while True:
                try:
                    round_config = self.round_configs.get(block=False)

                    # Check that the minimum allowed number of clients are connected
                    ready = self._check_nr_round_clients(round_config)
                    round_meta = {}

                    if ready:
                        if round_config['task'] == 'training':
                            tic = time.time()
                            round_meta = self.execute_training_round(round_config)
                            round_meta['time_exec_training'] = time.time() - \
                                tic
                            round_meta['status'] = "Success"
                            round_meta['name'] = self.server.id
                            self.server.tracer.set_round_combiner_data(round_meta)
                        elif round_config['task'] == 'validation' or round_config['task'] == 'inference':
                            self.execute_validation_round(round_config)
                        else:
                            self.server.report_status(
                                "ROUNDCONTROL: Round config contains unkown task type.", flush=True)
                    else:
                        round_meta = {}
                        round_meta['status'] = "Failed"
                        round_meta['reason'] = "Failed to meet client allocation requirements for this round config."
                        self.server.report_status(
                            "ROUNDCONTROL: {0}".format(round_meta['reason']), flush=True)

                    self.round_configs.task_done()
                except queue.Empty:
                    time.sleep(polling_interval)

        except (KeyboardInterrupt, SystemExit):
            pass



style = """
<style>
    /* Space out content a bit */
body {
  padding-top: 20px;
  padding-bottom: 20px;
}

/* Everything but the jumbotron gets side spacing for mobile first views */
.header,
.marketing,
.footer {
  padding-right: 15px;
  padding-left: 15px;
}

/* Custom page header */
.header {
  padding-bottom: 20px;
  border-bottom: 1px solid #e5e5e5;
}
.header h3 {
  margin-top: 0;
  margin-bottom: 0;
  line-height: 40px; /* Make the masthead heading the same height as the navigation */
}

/* Custom page footer */
.footer {
  padding-top: 19px;
  color: #777;
  border-top: 1px solid #e5e5e5;
}

/* Customize container */
@media (min-width: 768px) {
  .container {
    max-width: 730px;
  }
}
.container-narrow > hr {
  margin: 30px 0;
}

/* Main marketing message and sign up button */
.jumbotron {
  text-align: center;
  border-bottom: 1px solid #e5e5e5;
}
.jumbotron .btn {
  padding: 14px 24px;
  font-size: 21px;
}

/* Supporting marketing content */
.marketing {
  margin: 40px 0;
}
.marketing p + h4 {
  margin-top: 28px;
}

/* Responsive: Portrait tablets and up */
@media screen and (min-width: 768px) {
  /* Remove the padding we set earlier */
  .header,
  .marketing,
  .footer {
    padding-right: 0;
    padding-left: 0;
  }
  .header {
    margin-bottom: 30px;
  }
  .jumbotron {
    border-bottom: 0;
  }
}</style>
"""

page = """

<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <!-- The above 3 meta tags *must* come first in the head; any other head content must come *after* these tags -->
    <meta http-equiv="refresh" content="1">
    <meta name="description" content="">
    <meta name="author" content="">
    <link rel="icon" href="../../favicon.ico">
    <link rel="canonical" href="https://getbootstrap.com/docs/3.4/examples/jumbotron-narrow/">
<link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.5.0/css/bootstrap.min.css" integrity="sha384-9aIt2nRpC12Uk9gS9baDl411NQApFmC26EwAOH8WgZl5MYYxFfc+NcPb1dKGj7Sk" crossorigin="anonymous">

    <title>FEDn Client</title>

    {style}
    

    <!-- HTML5 shim and Respond.js for IE8 support of HTML5 elements and media queries -->
    <!--[if lt IE 9]>
      <script src="https://oss.maxcdn.com/html5shiv/3.7.3/html5shiv.min.js"></script>
      <script src="https://oss.maxcdn.com/respond/1.4.2/respond.min.js"></script>
    <![endif]-->
  </head>

  <body>

    <div class="container">
      <div class="header clearfix">
        <nav>
          <ul class="nav nav-pills pull-right">
            <li role="presentation" class="active">Status</li>
          </ul>
        </nav>
        <h3 class="text-muted">FEDn Client {client}</h3>
      </div>

      <div class="jumbotron">
        <p><a class="btn btn-lg btn-success" href="#" role="button">{state}</a></p>
      </div>

      <div class="row marketing">
        <div class="col-lg-12">
          <h4>Logs</h4>
          <div>
          {logs}
          </div> 
          
        </div>
      </div>

      <footer class="footer">
        <p>&copy; 2020 Scaleout Systems AB</p>
      </footer>

    </div> <!-- /container -->


    <!-- IE10 viewport hack for Surface/desktop Windows 8 bug -->
    <script src="https://code.jquery.com/jquery-3.5.1.slim.min.js" integrity="sha384-DfXdz2htPH0lsSSs5nCTpuj/zy4C+OGpamoFVy38MVBnE+IbbVYUew+OrCXaRkfj" crossorigin="anonymous"></script>
<script src="https://cdn.jsdelivr.net/npm/popper.js@1.16.0/dist/umd/popper.min.js" integrity="sha384-Q6E9RHvbIyZFJoft+2mJbHaEWldlvI9IOYy5n3zV9zzTtmI3UksdQRVvoxMfooAo" crossorigin="anonymous"></script>
<script src="https://stackpath.bootstrapcdn.com/bootstrap/4.5.0/js/bootstrap.min.js" integrity="sha384-OgVRvuATP1z7JjHLkuOU7Xw704+h835Lr+6QL9UvYjZE3Ipu6Tp75j7Bh/kR0JKI" crossorigin="anonymous"></script>

  </body>
</html>
"""  # noqa E501



import base64
import json
import queue
import re
import signal
import sys
import threading
import time
import uuid
from datetime import datetime, timedelta
from enum import Enum

import fedn.common.net.grpc.fedn_pb2 as fedn
import fedn.common.net.grpc.fedn_pb2_grpc as rpc
from fedn.common.net.grpc.server import Server
from fedn.common.storage.s3.s3repo import S3ModelRepository
from fedn.common.tracer.mongotracer import MongoTracer
from fedn.network.combiner.connect import ConnectorCombiner, Status
from fedn.network.combiner.modelservice import ModelService
from fedn.network.combiner.round import RoundController

VALID_NAME_REGEX = '^[a-zA-Z0-9_-]*$'


class Role(Enum):
    """ Enum for combiner roles. """
    WORKER = 1
    COMBINER = 2
    REDUCER = 3
    OTHER = 4


def role_to_proto_role(role):
    """ Convert a Role to a proto Role.

    :param role: the role to convert
    :type role: :class:`fedn.network.combiner.server.Role`
    :return: proto role
    :rtype: :class:`fedn.common.net.grpc.fedn_pb2.Role`
    """
    if role == Role.COMBINER:
        return fedn.COMBINER
    if role == Role.WORKER:
        return fedn.WORKER
    if role == Role.REDUCER:
        return fedn.REDUCER
    if role == Role.OTHER:
        return fedn.OTHER


class Combiner(rpc.CombinerServicer, rpc.ReducerServicer, rpc.ConnectorServicer, rpc.ControlServicer):
    """ Combiner gRPC server.

    :param config: configuration for the combiner
    :type config: dict
    """

    def __init__(self, config):
        """ Initialize Combiner server."""

        # Client queues
        self.clients = {}

        self.modelservice = ModelService()

        # Validate combiner name
        match = re.search(VALID_NAME_REGEX, config['name'])
        if not match:
            raise ValueError('Unallowed character in combiner name. Allowed characters: a-z, A-Z, 0-9, _, -.')

        self.id = config['name']
        self.role = Role.COMBINER
        self.max_clients = config['max_clients']

        # Connector to announce combiner to discover service (reducer)
        announce_client = ConnectorCombiner(host=config['discover_host'],
                                            port=config['discover_port'],
                                            myhost=config['host'],
                                            fqdn=config['fqdn'],
                                            myport=config['port'],
                                            token=config['token'],
                                            name=config['name'],
                                            secure=config['secure'],
                                            verify=config['verify'])

        response = None
        while True:
            # announce combiner to discover service
            status, response = announce_client.announce()
            if status == Status.TryAgain:
                print(response, flush=True)
                time.sleep(5)
                continue
            if status == Status.Assigned:
                announce_config = response
                print(
                    "COMBINER {0}: Announced successfully".format(self.id), flush=True)
                break
            if status == Status.UnAuthorized:
                print(response, flush=True)
                print("Status.UnAuthorized", flush=True)
                sys.exit("Exiting: Unauthorized")
            if status == Status.UnMatchedConfig:
                print(response, flush=True)
                print("Status.UnMatchedConfig", flush=True)
                sys.exit("Exiting: Missing config")

        cert = announce_config['certificate']
        key = announce_config['key']

        if announce_config['certificate']:
            cert = base64.b64decode(announce_config['certificate'])  # .decode('utf-8')
            key = base64.b64decode(announce_config['key'])  # .decode('utf-8')

        # Set up gRPC server configuration
        grpc_config = {'port': config['port'],
                       'secure': config['secure'],
                       'certificate': cert,
                       'key': key}

        # Set up model repository
        self.repository = S3ModelRepository(
            announce_config['storage']['storage_config'])

        # Create gRPC server
        self.server = Server(self, self.modelservice, grpc_config)

        # Set up tracer for statestore
        self.tracer = MongoTracer(
            announce_config['statestore']['mongo_config'], announce_config['statestore']['network_id'])

        # Set up round controller
        self.control = RoundController(config['aggregator'], self.repository, self, self.modelservice)

        # Start thread for round controller
        threading.Thread(target=self.control.run, daemon=True).start()

        # Start the gRPC server
        self.server.start()

    def __whoami(self, client, instance):
        """ Set the client id and role in a proto message.

        :param client: the client to set the id and role for
        :type client: :class:`fedn.common.net.grpc.fedn_pb2.Client`
        :param instance: the instance to get the id and role from
        :type instance: :class:`fedn.network.combiner.server.Combiner`
        :return: the client with id and role set
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.Client`
        """
        client.name = instance.id
        client.role = role_to_proto_role(instance.role)
        return client

    def report_status(self, msg, log_level=fedn.Status.INFO, type=None, request=None, flush=True):
        """ Report status of the combiner.

        :param msg: the message to report
        :type msg: str
        :param log_level: the log level to report at
        :type log_level: :class:`fedn.common.net.grpc.fedn_pb2.Status`
        :param type: the type of status to report
        :type type: :class:`fedn.common.net.grpc.fedn_pb2.Status.Type`
        :param request: the request to report status for
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.Request`
        :param flush: whether to flush the message to stdout
        :type flush: bool
        """
        print("{}:COMBINER({}):{} {}".format(datetime.now().strftime(
            '%Y-%m-%d %H:%M:%S'), self.id, log_level, msg), flush=flush)

    def request_model_update(self, config, clients=[]):
        """ Ask clients to update the current global model.

        :param config: the model configuration to send to clients
        :type config: dict
        :param clients: the clients to send the request to
        :type clients: list

        """

        request = fedn.ModelUpdateRequest()
        self.__whoami(request.sender, self)
        request.model_id = config['model_id']
        request.correlation_id = str(uuid.uuid4())
        request.timestamp = str(datetime.now())
        request.data = json.dumps(config)

        if len(clients) == 0:
            clients = self.get_active_trainers()

        for client in clients:
            request.receiver.name = client.name
            request.receiver.role = fedn.WORKER
            _ = self.SendModelUpdateRequest(request, self)
            # TODO: Check response

        print("COMBINER: Sent model update request for model {} to clients {}".format(
            request.model_id, clients), flush=True)

    def request_model_validation(self, model_id, config, clients=[]):
        """ Ask clients to validate the current global model.

        :param model_id: the model id to validate
        :type model_id: str
        :param config: the model configuration to send to clients
        :type config: dict
        :param clients: the clients to send the request to
        :type clients: list

        """

        request = fedn.ModelValidationRequest()
        self.__whoami(request.sender, self)
        request.model_id = model_id
        request.correlation_id = str(uuid.uuid4())
        request.timestamp = str(datetime.now())
        request.is_inference = (config['task'] == 'inference')

        if len(clients) == 0:
            clients = self.get_active_validators()

        for client in clients:
            request.receiver.name = client.name
            request.receiver.role = fedn.WORKER
            self.SendModelValidationRequest(request, self)

        print("COMBINER: Sent validation request for model {} to clients {}".format(
            model_id, clients), flush=True)

    def _list_clients(self, channel):
        """ List active clients on a channel.

        :param channel: the channel to list clients for, for example MODEL_UPDATE_REQUESTS
        :type channel: :class:`fedn.common.net.grpc.fedn_pb2.Channel`
        :return: the list of active clients
        :rtype: list
        """
        request = fedn.ListClientsRequest()
        self.__whoami(request.sender, self)
        request.channel = channel
        clients = self.ListActiveClients(request, self)
        return clients.client

    def get_active_trainers(self):
        """ Get a list of active trainers.

        :return: the list of active trainers
        :rtype: list
        """
        trainers = self._list_clients(fedn.Channel.MODEL_UPDATE_REQUESTS)
        return trainers

    def get_active_validators(self):
        """ Get a list of active validators.

        :return: the list of active validators
        :rtype: list
        """
        validators = self._list_clients(fedn.Channel.MODEL_VALIDATION_REQUESTS)
        return validators

    def nr_active_trainers(self):
        """ Get the number of active trainers.

        :return: the number of active trainers
        :rtype: int
        """
        return len(self.get_active_trainers())

    def nr_active_validators(self):
        """ Get the number of active validators.

        :return: the number of active validators
        :rtype: int
        """
        return len(self.get_active_validators())

    ####################################################################################################################

    def __join_client(self, client):
        """ Add a client to the list of active clients.

        :param client: the client to add
        :type client: :class:`fedn.common.net.grpc.fedn_pb2.Client`
        """
        if client.name not in self.clients.keys():
            self.clients[client.name] = {"lastseen": datetime.now()}

    def _subscribe_client_to_queue(self, client, queue_name):
        """ Subscribe a client to the queue.

        :param client: the client to subscribe
        :type client: :class:`fedn.common.net.grpc.fedn_pb2.Client`
        :param queue_name: the name of the queue to subscribe to
        :type queue_name: str
        """
        self.__join_client(client)
        if queue_name not in self.clients[client.name].keys():
            self.clients[client.name][queue_name] = queue.Queue()

    def __get_queue(self, client, queue_name):
        """ Get the queue for a client.

        :param client: the client to get the queue for
        :type client: :class:`fedn.common.net.grpc.fedn_pb2.Client`
        :param queue_name: the name of the queue to get
        :type queue_name: str
        :return: the queue
        :rtype: :class:`queue.Queue`

        :raises KeyError: if the queue does not exist
        """
        try:
            return self.clients[client.name][queue_name]
        except KeyError:
            raise

    def _send_request(self, request, queue_name):
        """ Send a request to a client.

        :param request: the request to send
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.Request`
        :param queue_name: the name of the queue to send the request to
        :type queue_name: str
        """
        self.__route_request_to_client(request, request.receiver, queue_name)

    def _broadcast_request(self, request, queue_name):
        """ Publish a request to all subscribed members.

        :param request: the request to send
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.Request`
        :param queue_name: the name of the queue to send the request to
        :type queue_name: str
        """
        active_clients = self._list_active_clients()
        for client in active_clients:
            self.clients[client.name][queue_name].put(request)

    def __route_request_to_client(self, request, client, queue_name):
        """ Route a request to a client.

        :param request: the request to send
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.Request`
        :param client: the client to send the request to
        :type client: :class:`fedn.common.net.grpc.fedn_pb2.Client`
        :param queue_name: the name of the queue to send the request to
        :type queue_name: str

        :raises Exception: if the request could not be routed, direct cause of KeyError in __get_queue
        """
        try:
            q = self.__get_queue(client, queue_name)
            q.put(request)
        except Exception:
            print("Failed to route request to client: {} {}",
                  request.receiver, queue_name)
            raise

    def _send_status(self, status):
        """ Report a status to tracer.

        :param status: the status to report
        :type status: :class:`fedn.common.net.grpc.fedn_pb2.Status`
        """

        self.tracer.report_status(status)

    def __register_heartbeat(self, client):
        """ Register a client if first time connecting. Update heartbeat timestamp.

        :param client: the client to register
        :type client: :class:`fedn.common.net.grpc.fedn_pb2.Client`
        """
        self.__join_client(client)
        self.clients[client.name]["lastseen"] = datetime.now()

    def flush_model_update_queue(self):
        """Clear the model update queue (aggregator). """

        q = self.control.aggregator.model_updates
        try:
            with q.mutex:
                q.queue.clear()
                q.all_tasks_done.notify_all()
                q.unfinished_tasks = 0
            return True
        except Exception:
            return False

    #####################################################################################################################

    # Control Service

    def Start(self, control: fedn.ControlRequest, context):
        """ Start a round of federated learning"

        :param control: the control request
        :type control: :class:`fedn.common.net.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.ControlResponse`
        """
        print("\nRECIEVED **START** from Controller {}\n".format(control.command), flush=True)

        config = {}
        for parameter in control.parameter:
            config.update({parameter.key: parameter.value})

        job_id = self.control.push_round_config(config)

        response = fedn.ControlResponse()
        p = response.parameter.add()
        p.key = "job_id"
        p.value = job_id

        return response

    # RPCs related to remote configuration of the server, round controller,
    # aggregator and their states.

    def Configure(self, control: fedn.ControlRequest, context):
        """ Configure the Combiner.

        :param control: the control request
        :type control: :class:`fedn.common.net.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.ControlResponse`
        """
        for parameter in control.parameter:
            setattr(self, parameter.key, parameter.value)

        response = fedn.ControlResponse()
        return response

    def FlushAggregationQueue(self, control: fedn.ControlRequest, context):
        """ Flush the queue.

        :param control: the control request
        :type control: :class:`fedn.common.net.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.ControlResponse`
        """

        status = self.flush_model_update_queue()

        response = fedn.ControlResponse()
        if status:
            response.message = 'Success'
        else:
            response.message = 'Failed'

        return response

    ##############################################################################

    def Stop(self, control: fedn.ControlRequest, context):
        """ TODO: Not yet implemented.

        :param control: the control request
        :type control: :class:`fedn.common.net.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.ControlResponse`
        """
        response = fedn.ControlResponse()
        print("\n RECIEVED **STOP** from Controller\n", flush=True)
        return response

    def Report(self, control: fedn.ControlRequest, context):
        """ Describe current state of the Combiner.

        :param control: the control request
        :type control: :class:`fedn.common.net.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.ControlResponse`
        """

        response = fedn.ControlResponse()
        self.report_status("\n RECIEVED **REPORT** from Controller\n",
                           log_level=fedn.Status.INFO)

        control_state = self.control.aggregator.get_state()
        self.report_status("Aggregator state: {}".format(control_state), log_level=fedn.Status.INFO)
        p = response.parameter.add()
        for key, value in control_state.items():
            p.key = str(key)
            p.value = str(value)

        active_trainers = self.get_active_trainers()
        p = response.parameter.add()
        p.key = "nr_active_trainers"
        p.value = str(len(active_trainers))

        active_validators = self.get_active_validators()
        p = response.parameter.add()
        p.key = "nr_active_validators"
        p.value = str(len(active_validators))

        active_trainers_ = self.get_active_trainers()
        active_trainers = []
        for client in active_trainers_:
            active_trainers.append(client)
        p = response.parameter.add()
        p.key = "active_trainers"
        p.value = str(active_trainers)

        active_validators_ = self.get_active_validators()
        active_validators = []
        for client in active_validators_:
            active_validators.append(client)
        p = response.parameter.add()
        p.key = "active_validators"
        p.value = str(active_validators)

        p = response.parameter.add()
        p.key = "nr_active_clients"
        p.value = str(len(active_trainers)+len(active_validators))

        p = response.parameter.add()
        p.key = "nr_unprocessed_compute_plans"
        p.value = str(self.control.round_configs.qsize())

        p = response.parameter.add()
        p.key = "name"
        p.value = str(self.id)

        return response

    #####################################################################################################################

    def SendStatus(self, status: fedn.Status, context):
        """ A client stream RPC endpoint that accepts status messages.

        :param status: the status message
        :type status: :class:`fedn.common.net.grpc.fedn_pb2.Status`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.Response`
        """

        self._send_status(status)

        response = fedn.Response()
        response.response = "Status received."
        return response

    def _list_subscribed_clients(self, queue_name):
        """ List all clients subscribed to a queue.

        :param queue_name: the name of the queue
        :type queue_name: str
        :return: a list of client names
        :rtype: list
        """
        subscribed_clients = []
        for name, client in self.clients.items():
            if queue_name in client.keys():
                subscribed_clients.append(name)
        return subscribed_clients

    def _list_active_clients(self, channel):
        """ List all clients that have sent a status message in the last 10 seconds.

        :param channel: the name of the channel
        :type channel: str
        :return: a list of client names
        :rtype: list
        """
        active_clients = []
        for client in self._list_subscribed_clients(channel):
            # This can break with different timezones.
            now = datetime.now()
            then = self.clients[client]["lastseen"]
            # TODO: move the heartbeat timeout to config.
            if (now - then) < timedelta(seconds=10):
                active_clients.append(client)
        return active_clients

    def _drop_inactive_clients(self):
        """ TODO: Not implemented. Clean up clients that have missed the heartbeat. """

    def ListActiveClients(self, request: fedn.ListClientsRequest, context):
        """ RPC endpoint that returns a ClientList containing the names of all active clients.
            An active client has sent a status message / responded to a heartbeat
            request in the last 10 seconds.

        :param request: the request
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.ListClientsRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the client list
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.ClientList`
        """
        clients = fedn.ClientList()
        active_clients = self._list_active_clients(request.channel)

        for client in active_clients:
            clients.client.append(fedn.Client(name=client, role=fedn.WORKER))
        return clients

    def AcceptingClients(self, request: fedn.ConnectionRequest, context):
        """ RPC endpoint that returns a ConnectionResponse indicating whether the server
        is accepting clients or not.

        :param request: the request (unused)
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.ConnectionRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.ConnectionResponse`
        """
        response = fedn.ConnectionResponse()
        active_clients = self._list_active_clients(
            fedn.Channel.MODEL_UPDATE_REQUESTS)

        try:
            requested = int(self.max_clients)
            if len(active_clients) >= requested:
                response.status = fedn.ConnectionStatus.NOT_ACCEPTING
                return response
            if len(active_clients) < requested:
                response.status = fedn.ConnectionStatus.ACCEPTING
                return response

        except Exception as e:
            print("Combiner not properly configured! {}".format(e), flush=True)
            raise

        response.status = fedn.ConnectionStatus.TRY_AGAIN_LATER
        return response

    def SendHeartbeat(self, heartbeat: fedn.Heartbeat, context):
        """ RPC that lets clients send a hearbeat, notifying the server that
            the client is available.

        :param heartbeat: the heartbeat
        :type heartbeat: :class:`fedn.common.net.grpc.fedn_pb2.Heartbeat`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.Response`
        """
        self.__register_heartbeat(heartbeat.sender)
        response = fedn.Response()
        response.sender.name = heartbeat.sender.name
        response.sender.role = heartbeat.sender.role
        response.response = "Heartbeat received"
        return response

    # Combiner Service

    def ModelUpdateStream(self, update, context):
        """ Model update stream RPC endpoint. Update status for client is connecting to stream.

        :param update: the update message
        :type update: :class:`fedn.common.net.grpc.fedn_pb2.ModelUpdate`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        """
        client = update.sender
        status = fedn.Status(
            status="Client {} connecting to ModelUpdateStream.".format(client.name))
        status.log_level = fedn.Status.INFO
        status.sender.name = self.id
        status.sender.role = role_to_proto_role(self.role)

        self._subscribe_client_to_queue(client, fedn.Channel.MODEL_UPDATES)
        q = self.__get_queue(client, fedn.Channel.MODEL_UPDATES)

        self._send_status(status)

        while context.is_active():
            try:
                yield q.get(timeout=1.0)
            except queue.Empty:
                pass

    def ModelUpdateRequestStream(self, response, context):
        """ A server stream RPC endpoint (Update model). Messages from client stream.

        :param response: the response
        :type response: :class:`fedn.common.net.grpc.fedn_pb2.ModelUpdateRequest`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        """

        client = response.sender
        metadata = context.invocation_metadata()
        if metadata:
            metadata = dict(metadata)
            print("\nClient connected: {}\n".format(metadata['client']), flush=True)

        status = fedn.Status(
            status="Client {} connecting to ModelUpdateRequestStream.".format(client.name))
        status.log_level = fedn.Status.INFO
        status.timestamp = str(datetime.now())

        self.__whoami(status.sender, self)

        self._subscribe_client_to_queue(
            client, fedn.Channel.MODEL_UPDATE_REQUESTS)
        q = self.__get_queue(client, fedn.Channel.MODEL_UPDATE_REQUESTS)

        self._send_status(status)

        self.tracer.update_client_status(client.name, "online")

        while context.is_active():
            try:
                yield q.get(timeout=1.0)
            except queue.Empty:
                pass

        self.tracer.update_client_status(client.name, "offline")

    def ModelValidationStream(self, update, context):
        """ Model validation stream RPC endpoint. Update status for client is connecting to stream.

        :param update: the update message
        :type update: :class:`fedn.common.net.grpc.fedn_pb2.ModelValidation`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        """
        client = update.sender
        status = fedn.Status(
            status="Client {} connecting to ModelValidationStream.".format(client.name))
        status.log_level = fedn.Status.INFO

        status.sender.name = self.id
        status.sender.role = role_to_proto_role(self.role)

        self._subscribe_client_to_queue(client, fedn.Channel.MODEL_VALIDATIONS)
        q = self.__get_queue(client, fedn.Channel.MODEL_VALIDATIONS)

        self._send_status(status)

        while context.is_active():
            try:
                yield q.get(timeout=1.0)
            except queue.Empty:
                pass

    def ModelValidationRequestStream(self, response, context):
        """ A server stream RPC endpoint (Validation). Messages from client stream.

        :param response: the response
        :type response: :class:`fedn.common.net.grpc.fedn_pb2.ModelValidationRequest`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        """

        client = response.sender
        status = fedn.Status(
            status="Client {} connecting to ModelValidationRequestStream.".format(client.name))
        status.log_level = fedn.Status.INFO
        status.sender.name = self.id
        status.sender.role = role_to_proto_role(self.role)
        status.timestamp = str(datetime.now())

        self._subscribe_client_to_queue(
            client, fedn.Channel.MODEL_VALIDATION_REQUESTS)
        q = self.__get_queue(client, fedn.Channel.MODEL_VALIDATION_REQUESTS)

        self._send_status(status)

        while context.is_active():
            try:
                yield q.get(timeout=1.0)
            except queue.Empty:
                pass

    def SendModelUpdateRequest(self, request, context):
        """ Send a model update request.

        :param request: the request
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.ModelUpdateRequest`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.Response`
        """
        self._send_request(request, fedn.Channel.MODEL_UPDATE_REQUESTS)

        response = fedn.Response()
        response.response = "RECEIVED ModelUpdateRequest from client {}".format(
            request.sender.name)
        return response  # TODO Fill later

    def SendModelUpdate(self, request, context):
        """ Send a model update response.

        :param request: the request
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.ModelUpdate`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.Response`
        """
        self.control.aggregator.on_model_update(request)

        response = fedn.Response()
        response.response = "RECEIVED ModelUpdate {} from client  {}".format(
            response, response.sender.name)
        return response  # TODO Fill later

    def SendModelValidationRequest(self, request, context):
        """ Send a model validation request.

        :param request: the request
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.ModelValidationRequest`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.Response`
        """
        self._send_request(request, fedn.Channel.MODEL_VALIDATION_REQUESTS)

        response = fedn.Response()
        response.response = "RECEIVED ModelValidationRequest from client {}".format(
            request.sender.name)
        return response  # TODO Fill later

    def register_model_validation(self, validation):
        """Register a model validation.

        :param validation: the model validation
        :type validation: :class:`fedn.common.net.grpc.fedn_pb2.ModelValidation`
        """

        self.tracer.report_validation(validation)

    def SendModelValidation(self, request, context):
        """ Send a model validation response.

        :param request: the request
        :type request: :class:`fedn.common.net.grpc.fedn_pb2.ModelValidation`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.common.net.grpc.fedn_pb2.Response`
        """
        self.report_status("Recieved ModelValidation from {}".format(request.sender.name),
                           log_level=fedn.Status.INFO)

        self.register_model_validation(request)

        response = fedn.Response()
        response.response = "RECEIVED ModelValidation {} from client  {}".format(
            response, response.sender.name)
        return response

    ####################################################################################################################

    def run(self):
        """ Start the server."""

        print("COMBINER: {} started, ready for requests. ".format(
            self.id), flush=True)
        try:
            while True:
                signal.pause()
        except (KeyboardInterrupt, SystemExit):
            pass
        self.server.stop()



import base64
import copy
import datetime
import json
import os
import re
import threading
from io import BytesIO
from threading import Lock

import jwt
import pandas as pd
from bokeh.embed import json_item
from bson import json_util
from flask import (Flask, abort, flash, jsonify, make_response, redirect,
                   render_template, request, send_file, send_from_directory,
                   url_for)
from werkzeug.utils import secure_filename

from fedn.common.tracer.mongotracer import MongoTracer
from fedn.network.combiner.interfaces import CombinerInterface
from fedn.network.dashboard.plots import Plot
from fedn.network.state import ReducerState, ReducerStateToString
from fedn.utils.checksum import sha

UPLOAD_FOLDER = "/app/client/package/"
ALLOWED_EXTENSIONS = {"gz", "bz2", "tar", "zip", "tgz"}


def allowed_file(filename):
    """

    :param filename:
    :return:
    """
    return (
        "." in filename
        and filename.rsplit(".", 1)[1].lower() in ALLOWED_EXTENSIONS
    )


def encode_auth_token(secret_key):
    """Generates the Auth Token
    :return: string
    """
    try:
        payload = {
            "exp": datetime.datetime.utcnow()
            + datetime.timedelta(days=90, seconds=0),
            "iat": datetime.datetime.utcnow(),
            "status": "Success",
        }
        token = jwt.encode(payload, secret_key, algorithm="HS256")
        print(
            "\n\n\nSECURE MODE ENABLED, USE TOKEN TO ACCESS REDUCER: **** {} ****\n\n\n".format(
                token
            )
        )
        return token
    except Exception as e:
        return e


def decode_auth_token(auth_token, secret):
    """Decodes the auth token
    :param auth_token:
    :return: string
    """
    try:
        payload = jwt.decode(auth_token, secret, algorithms=["HS256"])
        return payload["status"]
    except jwt.ExpiredSignatureError as e:
        print(e)
        return "Token has expired."
    except jwt.InvalidTokenError as e:
        print(e)
        return "Invalid token."


class ReducerRestService:
    """ """

    def __init__(self, config, control, statestore, certificate_manager):
        print("config object!: \n\n\n\n{}".format(config))
        if config["host"]:
            self.host = config["host"]
        else:
            self.host = None

        self.name = config["name"]

        self.port = config["port"]
        self.network_id = config["name"] + "-network"

        if "token" in config.keys():
            self.token_auth_enabled = True
        else:
            self.token_auth_enabled = False

        if "secret_key" in config.keys():
            self.SECRET_KEY = config["secret_key"]
        else:
            self.SECRET_KEY = None

        if "use_ssl" in config.keys():
            self.use_ssl = config["use_ssl"]

        self.remote_compute_package = config["remote_compute_package"]
        if self.remote_compute_package:
            self.package = "remote"
        else:
            self.package = "local"

        self.control = control
        self.statestore = statestore
        self.certificate_manager = certificate_manager
        self.current_compute_context = None

    def to_dict(self):
        """

        :return:
        """
        data = {"name": self.name}
        return data

    def check_compute_package(self):
        """Check if the compute package has been configured,
        if remote compute context is set to False, True will be returned

        :return: True if configured
        :rtype: bool
        """
        if not self.remote_compute_package:
            return True

        if not self.control.get_compute_package():
            return False
        else:
            return True

    def check_initial_model(self):
        """Check if initial model (seed model) has been configured

        :return: True if configured, else False
        :rtype: bool
        """

        if self.statestore.get_latest_model():
            return True
        else:
            return False

    def check_configured_response(self):
        """Check if everything has been configured for client to connect,
        return response if not.

        :return: Reponse with message if not configured, else None
        :rtype: json
        """
        if self.control.state() == ReducerState.setup:
            return jsonify(
                {
                    "status": "retry",
                    "package": self.package,
                    "msg": "Controller is not configured.",
                }
            )

        if not self.check_compute_package():
            return jsonify(
                {
                    "status": "retry",
                    "package": self.package,
                    "msg": "Compute package is not configured. Please upload the compute package.",
                }
            )

        if not self.check_initial_model():
            return jsonify(
                {
                    "status": "retry",
                    "package": self.package,
                    "msg": "Initial model is not configured. Please upload the model.",
                }
            )

        if not self.control.idle():
            return jsonify(
                {
                    "status": "retry",
                    "package": self.package,
                    "msg": "Conroller is not in idle state, try again later. ",
                }
            )
        return None

    def check_configured(self):
        """Check if compute package has been configured and that and that the
        state of the ReducerControl is not in setup otherwise render setup template.
        Check if initial model has been configured, otherwise render setup_model template.
        :return: Rendered html template or None
        """
        if not self.check_compute_package():
            return render_template(
                "setup.html",
                client=self.name,
                state=ReducerStateToString(self.control.state()),
                logs=None,
                refresh=False,
                message="Please set the compute package",
            )

        if self.control.state() == ReducerState.setup:
            return render_template(
                "setup.html",
                client=self.name,
                state=ReducerStateToString(self.control.state()),
                logs=None,
                refresh=True,
                message="Warning. Reducer is not base-configured. please do so with config file.",
            )

        if not self.check_initial_model():
            return render_template(
                "setup_model.html", message="Please set the initial model."
            )

        return None

    def authorize(self, r, secret):
        """Authorize client token

        :param r: Request
        :type r: [type]
        :param token: Token to verify against
        :type token: string
        """
        try:
            # Get token
            if "Authorization" in r.headers:  # header auth
                request_token = r.headers.get("Authorization").split()[1]
            elif "token" in r.args:  # args auth
                request_token = str(r.args.get("token"))
            elif "fedn_token" in r.cookies:
                request_token = r.cookies.get("fedn_token")
            else:  # no token provided
                print("Authorization failed. No token provided.", flush=True)
                abort(401)

            # Log token and secret
            print(
                f"Secret: {secret}. Request token: {request_token}.",
                flush=True,
            )

            # Authenticate
            status = decode_auth_token(request_token, secret)
            if status == "Success":
                return True
            else:
                print(
                    'Authorization failed. Status: "{}"'.format(status),
                    flush=True,
                )
                abort(401)
        except Exception as e:
            print(
                'Authorization failed. Expection encountered: "{}".'.format(e),
                flush=True,
            )
            abort(401)

    def run(self):
        """

        :return:
        """
        app = Flask(__name__)

        app.config["UPLOAD_FOLDER"] = UPLOAD_FOLDER
        app.config["SECRET_KEY"] = self.SECRET_KEY

        @app.route("/")
        def index():
            """

            :return:
            """
            # Token auth
            if self.token_auth_enabled:
                self.authorize(request, app.config.get("SECRET_KEY"))

            # Render template
            not_configured_template = self.check_configured()
            if not_configured_template:
                template = not_configured_template
            else:
                events = self.control.get_events()
                message = request.args.get("message", None)
                message_type = request.args.get("message_type", None)
                template = render_template(
                    "events.html",
                    client=self.name,
                    state=ReducerStateToString(self.control.state()),
                    events=events,
                    logs=None,
                    refresh=True,
                    configured=True,
                    message=message,
                    message_type=message_type,
                )

            # Set token cookie in response if needed
            response = make_response(template)
            if "token" in request.args:  # args auth
                response.set_cookie("fedn_token", str(request.args["token"]))

            # Return response
            return response

        @app.route("/status")
        def status():
            """

            :return:
            """
            return {"state": ReducerStateToString(self.control.state())}

        @app.route("/netgraph")
        def netgraph():
            """
            Creates nodes and edges for network graph

            :return: nodes and edges as keys
            :rtype: dict
            """
            result = {"nodes": [], "edges": []}

            result["nodes"].append(
                {
                    "id": "reducer",
                    "label": "Reducer",
                    "role": "reducer",
                    "status": "active",
                    "name": "reducer",  # TODO: get real host name
                    "type": "reducer",
                }
            )

            combiner_info = combiner_status()
            client_info = client_status()

            if len(combiner_info) < 1:
                return result

            for combiner in combiner_info:
                print("combiner info {}".format(combiner_info), flush=True)
                try:
                    result["nodes"].append(
                        {
                            "id": combiner["name"],  # "n{}".format(count),
                            "label": "Combiner ({} clients)".format(
                                combiner["nr_active_clients"]
                            ),
                            "role": "combiner",
                            "status": "active",  # TODO: Hard-coded, combiner_info does not contain status
                            "name": combiner["name"],
                            "type": "combiner",
                        }
                    )
                except Exception as err:
                    print(err)

            for client in client_info["active_clients"]:
                try:
                    if client["status"] != "offline":
                        result["nodes"].append(
                            {
                                "id": str(client["_id"]),
                                "label": "Client",
                                "role": client["role"],
                                "status": client["status"],
                                "name": client["name"],
                                "combiner": client["combiner"],
                                "type": "client",
                            }
                        )
                except Exception as err:
                    print(err)

            count = 0
            for node in result["nodes"]:
                try:
                    if node["type"] == "combiner":
                        result["edges"].append(
                            {
                                "id": "e{}".format(count),
                                "source": node["id"],
                                "target": "reducer",
                            }
                        )
                    elif node["type"] == "client":
                        result["edges"].append(
                            {
                                "id": "e{}".format(count),
                                "source": node["combiner"],
                                "target": node["id"],
                            }
                        )
                except Exception:
                    pass
                count = count + 1
            return result

        @app.route("/networkgraph")
        def network_graph():
            try:
                plot = Plot(self.control.statestore)
                result = netgraph()
                df_nodes = pd.DataFrame(result["nodes"])
                df_edges = pd.DataFrame(result["edges"])
                graph = plot.make_netgraph_plot(df_edges, df_nodes)
                return json.dumps(json_item(graph, "myplot"))
            except Exception:
                raise
                # return ''

        @app.route("/events")
        def events():
            """

            :return:
            """

            response = self.control.get_events()
            events = []

            result = response["result"]

            for evt in result:
                events.append(evt)

            return jsonify({"result": events, "count": response["count"]})

            json_docs = []
            for doc in self.control.get_events():
                json_doc = json.dumps(doc, default=json_util.default)
                json_docs.append(json_doc)

            json_docs.reverse()

            return {"events": json_docs}

        @app.route("/add")
        def add():
            """Add a combiner to the network."""
            print("Adding combiner to network:", flush=True)
            if self.token_auth_enabled:
                self.authorize(request, app.config.get("SECRET_KEY"))
            if self.control.state() == ReducerState.setup:
                return jsonify({"status": "retry"})

            name = request.args.get("name", None)
            address = str(request.args.get("address", None))
            fqdn = str(request.args.get("fqdn", None))
            port = request.args.get("port", None)
            secure_grpc = request.args.get("secure", None)

            if (
                port is None
                or address is None
                or name is None
                or secure_grpc is None
            ):
                return "Please specify correct parameters."

            # Try to retrieve combiner from db
            combiner = self.control.network.get_combiner(name)
            if not combiner:
                if secure_grpc == "True":
                    certificate, key = self.certificate_manager.get_or_create(
                        address
                    ).get_keypair_raw()
                    _ = base64.b64encode(certificate)
                    _ = base64.b64encode(key)

                else:
                    certificate = None
                    key = None

                combiner = CombinerInterface(
                    self,
                    name=name,
                    address=address,
                    fqdn=fqdn,
                    port=port,
                    certificate=copy.deepcopy(certificate),
                    key=copy.deepcopy(key),
                    ip=request.remote_addr,
                )

                self.control.network.add_combiner(combiner)

            combiner = self.control.network.get_combiner(name)

            ret = {
                "status": "added",
                "storage": self.control.statestore.get_storage_backend(),
                "statestore": self.control.statestore.get_config(),
                "certificate": combiner.get_certificate(),
                "key": combiner.get_key(),
            }

            return jsonify(ret)

        @app.route("/eula", methods=["GET", "POST"])
        def eula():
            """

            :return:
            """
            for r in request.headers:
                print("header contains: {}".format(r), flush=True)

            return render_template("eula.html", configured=True)

        @app.route("/models", methods=["GET", "POST"])
        def models():
            """

            :return:
            """
            # Token auth
            if self.token_auth_enabled:
                self.authorize(request, app.config.get("SECRET_KEY"))

            if request.method == "POST":
                # upload seed file
                uploaded_seed = request.files["seed"]
                if uploaded_seed:
                    a = BytesIO()
                    a.seek(0, 0)
                    uploaded_seed.seek(0)
                    a.write(uploaded_seed.read())
                    helper = self.control.get_helper()
                    a.seek(0)
                    model = helper.load(a)
                    self.control.commit(uploaded_seed.filename, model)
            else:
                not_configured = self.check_configured()
                if not_configured:
                    return not_configured

                plot = Plot(self.control.statestore)
                try:
                    valid_metrics = plot.fetch_valid_metrics()
                    box_plot = plot.create_box_plot(valid_metrics[0])
                except Exception as e:
                    valid_metrics = None
                    box_plot = None
                    print(e, flush=True)

                h_latest_model_id = self.statestore.get_latest_model()

                model_info = self.control.get_model_info()
                return render_template(
                    "models.html",
                    box_plot=box_plot,
                    metrics=valid_metrics,
                    h_latest_model_id=h_latest_model_id,
                    seed=True,
                    model_info=model_info,
                    configured=True,
                )

            seed = True
            return redirect(url_for("models", seed=seed))

        @app.route("/delete_model_trail", methods=["GET", "POST"])
        def delete_model_trail():
            """

            :return:
            """
            if request.method == "POST":
                statestore_config = self.control.statestore.get_config()
                self.tracer = MongoTracer(
                    statestore_config["mongo_config"],
                    statestore_config["network_id"],
                )
                try:
                    self.control.drop_models()
                except Exception:
                    pass

                # drop objects in minio
                self.control.delete_bucket_objects()
                return redirect(url_for("models"))
            seed = True
            return redirect(url_for("models", seed=seed))

        @app.route("/drop_control", methods=["GET", "POST"])
        def drop_control():
            """

            :return:
            """
            if request.method == "POST":
                self.control.statestore.drop_control()
                return redirect(url_for("control"))
            return redirect(url_for("control"))

        # http://localhost:8090/control?rounds=4&model_id=879fa112-c861-4cb1-a25d-775153e5b548
        @app.route("/control", methods=["GET", "POST"])
        def control():
            """Main page for round control. Configure, start and stop training sessions."""
            # Token auth
            if self.token_auth_enabled:
                self.authorize(request, app.config.get("SECRET_KEY"))

            not_configured = self.check_configured()
            if not_configured:
                return not_configured

            state = ReducerStateToString(self.control.state())
            refresh = True

            if self.remote_compute_package:
                try:
                    self.current_compute_context = (
                        self.control.get_compute_package_name()
                    )
                except Exception:
                    self.current_compute_context = None
            else:
                self.current_compute_context = "None:Local"
            if self.control.state() == ReducerState.monitoring:
                return redirect(
                    url_for(
                        "index",
                        state=state,
                        refresh=refresh,
                        message="Reducer is in monitoring state",
                    )
                )

            if request.method == "POST":
                # Get session configuration
                round_timeout = float(request.form.get("timeout", 180))
                buffer_size = int(request.form.get("buffer_size", -1))
                rounds = int(request.form.get("rounds", 1))
                delete_models = request.form.get("delete_models", True)
                task = request.form.get("task", "")
                clients_required = request.form.get("clients_required", 1)
                clients_requested = request.form.get("clients_requested", 8)

                # checking if there are enough clients connected to start!
                clients_available = 0
                for combiner in self.control.network.get_combiners():
                    try:
                        combiner_state = combiner.report()
                        nac = combiner_state["nr_active_clients"]
                        clients_available = clients_available + int(nac)
                    except Exception:
                        pass

                if clients_available < clients_required:
                    return redirect(
                        url_for(
                            "index",
                            state=state,
                            message="Not enough clients available to start rounds! "
                            "check combiner client capacity",
                            message_type="warning",
                        )
                    )

                validate = request.form.get("validate", False)
                if validate == "False":
                    validate = False
                helper_type = request.form.get("helper", "keras")
                # self.control.statestore.set_framework(helper_type)

                latest_model_id = self.statestore.get_latest_model()

                config = {
                    "round_timeout": round_timeout,
                    "buffer_size": buffer_size,
                    "model_id": latest_model_id,
                    "rounds": rounds,
                    "delete_models_storage": delete_models,
                    "clients_required": clients_required,
                    "clients_requested": clients_requested,
                    "task": task,
                    "validate": validate,
                    "helper_type": helper_type,
                }

                threading.Thread(
                    target=self.control.session, args=(config,)
                ).start()

                return redirect(
                    url_for(
                        "index",
                        state=state,
                        refresh=refresh,
                        message="Sent execution plan.",
                        message_type="SUCCESS",
                    )
                )

            else:
                seed_model_id = None
                latest_model_id = None
                try:
                    seed_model_id = self.statestore.get_initial_model()
                    latest_model_id = self.statestore.get_latest_model()
                except Exception:
                    pass

                return render_template(
                    "index.html",
                    latest_model_id=latest_model_id,
                    compute_package=self.current_compute_context,
                    seed_model_id=seed_model_id,
                    helper=self.control.statestore.get_helper(),
                    validate=True,
                    configured=True,
                )

        @app.route("/assign")
        def assign():
            """Handle client assignment requests."""

            if self.token_auth_enabled:
                self.authorize(request, app.config.get("SECRET_KEY"))

            response = self.check_configured_response()

            if response:
                return response

            name = request.args.get("name", None)
            combiner_preferred = request.args.get("combiner", None)

            if combiner_preferred:
                combiner = self.control.network.get_combiner(
                    combiner_preferred
                )
            else:
                combiner = self.control.network.find_available_combiner()

            if combiner is None:
                return jsonify(
                    {
                        "status": "retry",
                        "package": self.package,
                        "msg": "Failed to assign to a combiner, try again later.",
                    }
                )

            client = {
                "name": name,
                "combiner_preferred": combiner_preferred,
                "combiner": combiner.name,
                "ip": request.remote_addr,
                "status": "available",
            }

            # Add client to database
            self.control.network.add_client(client)

            # Return connection information to client
            if combiner.certificate:
                cert_b64 = base64.b64encode(combiner.certificate)
                cert = str(cert_b64).split("'")[1]
            else:
                cert = None

            response = {
                "status": "assigned",
                "host": combiner.address,
                "fqdn": combiner.fqdn,
                "package": self.package,
                "ip": combiner.ip,
                "port": combiner.port,
                "certificate": cert,
                "model_type": self.control.statestore.get_helper(),
            }

            return jsonify(response)

        def combiner_status():
            """Get current status reports from all combiners registered in the network.

            :return:
            """
            combiner_info = []
            for combiner in self.control.network.get_combiners():
                try:
                    report = combiner.report()
                    combiner_info.append(report)
                except Exception:
                    pass
            return combiner_info

        def client_status():
            """
            Get current status of clients (available) from DB compared with client status from all combiners,
            update client status to DB and add their roles.
            """
            client_info = self.control.network.get_client_info()
            combiner_info = combiner_status()
            try:
                all_active_trainers = []
                all_active_validators = []

                for client in combiner_info:
                    active_trainers_str = client["active_trainers"]
                    active_validators_str = client["active_validators"]
                    active_trainers_str = re.sub(
                        "[^a-zA-Z0-9-:\n\.]", "", active_trainers_str  # noqa: W605
                    ).replace(
                        "name:", " "
                    )
                    active_validators_str = re.sub(
                        "[^a-zA-Z0-9-:\n\.]", "", active_validators_str  # noqa: W605
                    ).replace(
                        "name:", " "
                    )
                    all_active_trainers.extend(
                        " ".join(active_trainers_str.split(" ")).split()
                    )
                    all_active_validators.extend(
                        " ".join(active_validators_str.split(" ")).split()
                    )

                active_trainers_list = [
                    client
                    for client in client_info
                    if client["name"] in all_active_trainers
                ]
                active_validators_list = [
                    cl
                    for cl in client_info
                    if cl["name"] in all_active_validators
                ]
                all_clients = [cl for cl in client_info]

                for client in all_clients:
                    status = "offline"
                    role = "None"
                    self.control.network.update_client_data(
                        client, status, role
                    )

                all_active_clients = (
                    active_validators_list + active_trainers_list
                )
                for client in all_active_clients:
                    status = "active"
                    if (
                        client in active_trainers_list
                        and client in active_validators_list
                    ):
                        role = "trainer-validator"
                    elif client in active_trainers_list:
                        role = "trainer"
                    elif client in active_validators_list:
                        role = "validator"
                    else:
                        role = "unknown"
                    self.control.network.update_client_data(
                        client, status, role
                    )

                return {
                    "active_clients": all_clients,
                    "active_trainers": active_trainers_list,
                    "active_validators": active_validators_list,
                }
            except Exception:
                pass

            return {
                "active_clients": [],
                "active_trainers": [],
                "active_validators": [],
            }

        @app.route("/metric_type", methods=["GET", "POST"])
        def change_features():
            """

            :return:
            """
            feature = request.args["selected"]
            plot = Plot(self.control.statestore)
            graphJSON = plot.create_box_plot(feature)
            return graphJSON

        @app.route("/dashboard")
        def dashboard():
            """

            :return:
            """
            # Token auth
            if self.token_auth_enabled:
                self.authorize(request, app.config.get("SECRET_KEY"))

            not_configured = self.check_configured()
            if not_configured:
                return not_configured

            plot = Plot(self.control.statestore)
            combiners_plot = plot.create_combiner_plot()

            timeline_plot = None
            table_plot = None
            clients_plot = plot.create_client_plot()
            client_histogram_plot = plot.create_client_histogram_plot()

            return render_template(
                "dashboard.html",
                show_plot=True,
                table_plot=table_plot,
                timeline_plot=timeline_plot,
                clients_plot=clients_plot,
                client_histogram_plot=client_histogram_plot,
                combiners_plot=combiners_plot,
                configured=True,
            )

        @app.route("/network")
        def network():
            """

            :return:
            """
            # Token auth
            if self.token_auth_enabled:
                self.authorize(request, app.config.get("SECRET_KEY"))

            not_configured = self.check_configured()
            if not_configured:
                return not_configured
            plot = Plot(self.control.statestore)
            round_time_plot = plot.create_round_plot()
            mem_cpu_plot = plot.create_cpu_plot()
            combiner_info = combiner_status()
            active_clients = client_status()
            # print(combiner_info, flush=True)
            return render_template(
                "network.html",
                network_plot=True,
                round_time_plot=round_time_plot,
                mem_cpu_plot=mem_cpu_plot,
                combiner_info=combiner_info,
                active_clients=active_clients["active_clients"],
                active_trainers=active_clients["active_trainers"],
                active_validators=active_clients["active_validators"],
                configured=True,
            )

        @app.route("/config/download", methods=["GET"])
        def config_download():
            """

            :return:
            """
            chk_string = ""
            name = self.control.get_compute_package_name()
            if name is None or name == "":
                chk_string = ""
            else:
                file_path = os.path.join(UPLOAD_FOLDER, name)
                print("trying to get {}".format(file_path))

                try:
                    sum = str(sha(file_path))
                except FileNotFoundError:
                    sum = ""
                chk_string = "checksum: {}".format(sum)

            network_id = self.network_id
            discover_host = self.name
            discover_port = self.port
            ctx = """network_id: {network_id}
discover_host: {discover_host}
discover_port: {discover_port}
{chk_string}""".format(
                network_id=network_id,
                discover_host=discover_host,
                discover_port=discover_port,
                chk_string=chk_string,
            )

            obj = BytesIO()
            obj.write(ctx.encode("UTF-8"))
            obj.seek(0)
            return send_file(
                obj,
                as_attachment=True,
                download_name="client.yaml",
                mimetype="application/x-yaml",
            )

        @app.route("/context", methods=["GET", "POST"])
        def context():
            """

            :return:
            """
            # Token auth
            if self.token_auth_enabled:
                self.authorize(request, app.config.get("SECRET_KEY"))

            # if reset is not empty then allow context re-set
            reset = request.args.get("reset", None)
            if reset:
                return render_template("context.html")

            if request.method == "POST":
                if "file" not in request.files:
                    flash("No file part")
                    return redirect(url_for("context"))

                file = request.files["file"]
                helper_type = request.form.get("helper", "kerashelper")
                # if user does not select file, browser also
                # submit an empty part without filename
                if file.filename == "":
                    flash("No selected file")
                    return redirect(url_for("context"))

                if file and allowed_file(file.filename):
                    filename = secure_filename(file.filename)
                    file_path = os.path.join(
                        app.config["UPLOAD_FOLDER"], filename
                    )
                    file.save(file_path)

                    if (
                        self.control.state() == ReducerState.instructing
                        or self.control.state() == ReducerState.monitoring
                    ):
                        return "Not allowed to change context while execution is ongoing."

                    self.control.set_compute_package(filename, file_path)
                    self.control.statestore.set_helper(helper_type)
                    return redirect(url_for("control"))

            name = request.args.get("name", "")

            if name == "":
                name = self.control.get_compute_package_name()
                if name is None or name == "":
                    return render_template("context.html")

            # There is a potential race condition here, if one client requests a package and at
            # the same time another one triggers a fetch from Minio and writes to disk.
            try:
                mutex = Lock()
                mutex.acquire()
                return send_from_directory(
                    app.config["UPLOAD_FOLDER"], name, as_attachment=True
                )
            except Exception:
                try:
                    data = self.control.get_compute_package(name)
                    file_path = os.path.join(app.config["UPLOAD_FOLDER"], name)
                    with open(file_path, "wb") as fh:
                        fh.write(data)
                    return send_from_directory(
                        app.config["UPLOAD_FOLDER"], name, as_attachment=True
                    )
                except Exception:
                    raise
            finally:
                mutex.release()

            return render_template("context.html")

        @app.route("/checksum", methods=["GET", "POST"])
        def checksum():
            """

            :return:
            """
            # sum = ''
            name = request.args.get("name", None)
            if name == "" or name is None:
                name = self.control.get_compute_package_name()
                if name is None or name == "":
                    return jsonify({})

            file_path = os.path.join(UPLOAD_FOLDER, name)
            print("trying to get {}".format(file_path))

            try:
                sum = str(sha(file_path))
            except FileNotFoundError:
                sum = ""

            data = {"checksum": sum}

            return jsonify(data)

        @app.route("/infer", methods=["POST"])
        def infer():
            """

            :return:
            """
            # Token auth
            if self.token_auth_enabled:
                self.authorize(request, app.config.get("SECRET_KEY"))

            # Check configured
            not_configured = self.check_configured()
            if not_configured:
                return not_configured

            # Check compute context
            if self.remote_compute_context:
                try:
                    self.current_compute_context = (
                        self.control.get_compute_package()
                    )
                except Exception as e:
                    print(e, flush=True)
                    self.current_compute_context = None
            else:
                self.current_compute_context = "None:Local"

            # Redirect if in monitoring state
            if self.control.state() == ReducerState.monitoring:
                return redirect(
                    url_for(
                        "index",
                        state=ReducerStateToString(self.control.state()),
                        refresh=True,
                        message="Reducer is in monitoring state",
                    )
                )

            # POST params
            timeout = int(request.form.get("timeout", 180))
            helper_type = request.form.get("helper", "keras")
            clients_required = request.form.get("clients_required", 1)
            clients_requested = request.form.get("clients_requested", 8)

            # Start inference request
            config = {
                "round_timeout": timeout,
                "model_id": self.statestore.get_latest_model(),
                "clients_required": clients_required,
                "clients_requested": clients_requested,
                "task": "inference",
                "helper_type": helper_type,
            }
            threading.Thread(
                target=self.control.infer_instruct, args=(config,)
            ).start()

            # Redirect
            return redirect(
                url_for(
                    "index",
                    state=ReducerStateToString(self.control.state()),
                    refresh=True,
                    message="Sent execution plan (inference).",
                    message_type="SUCCESS",
                )
            )

        if not self.host:
            bind = "0.0.0.0"
        else:
            bind = self.host

        app.run(host=bind, port=self.port)

        return app



import unittest
from unittest.mock import patch

from fedn.network.dashboard.restservice import ReducerRestService


class TestInit(unittest.TestCase):

    @patch('fedn.clients.reducer.control.ReducerControl')
    def test_discover_host(self, mock_control):
        CONFIG = {
            'discover_host': 'TEST_HOST',
            'name': 'TEST_NAME',
            'discover_port': 1111,
            'token': False,
            'remote_compute_context': True
        }
        restservice = ReducerRestService(CONFIG, mock_control, None)
        self.assertEqual(restservice.name, 'TEST_HOST')
        self.assertEqual(restservice.network_id, 'TEST_NAME-network')

    @patch('fedn.clients.reducer.control.ReducerControl')
    def test_name(self, mock_control):
        CONFIG = {
            'discover_host': None,
            'name': 'TEST_NAME',
            'discover_port': 1111,
            'token': False,
            'remote_compute_context': True
        }
        restservice = ReducerRestService(CONFIG, mock_control, None)
        self.assertEqual(restservice.name, 'TEST_NAME')

    @patch('fedn.clients.reducer.control.ReducerControl')
    def test_network_id(self, mock_control):
        CONFIG = {
            'discover_host': 'TEST_HOST',
            'name': 'TEST_NAME',
            'discover_port': 1111,
            'token': False,
            'remote_compute_context': True
        }
        restservice = ReducerRestService(CONFIG, mock_control, None)
        self.assertEqual(restservice.network_id, 'TEST_NAME-network')


class TestChecks(unittest.TestCase):
    @patch('fedn.clients.reducer.control.ReducerControl')
    def setUp(self, mock_control):
        CONFIG = {
            'discover_host': 'TEST_HOST',
            'name': 'TEST_NAME',
            'discover_port': 1111,
            'token': False,
            'remote_compute_context': True
        }

        self.restservice = ReducerRestService(CONFIG, mock_control, None)

    def test_check_compute_package(self):

        self.restservice.control.get_compute_context.return_value = {
            'NOT': 'NONE'}
        retval = self.restservice.check_compute_context()
        self.assertTrue(retval)

        self.restservice.control.get_compute_context.return_value = None
        retval = self.restservice.check_compute_context()
        self.assertFalse(retval)

        self.restservice.control.get_compute_context.return_value = ''
        retval = self.restservice.check_compute_context()
        self.assertFalse(retval)

        self.restservice.remote_compute_context = False
        retval = self.restservice.check_compute_context()
        self.assertTrue(retval)

    def test_check_initial_model(self):

        self.restservice.statestore.get_latest_model.return_value = 'model-uid'
        retval = self.restservice.check_initial_model()
        self.assertTrue(retval)

        self.restservice.statestore.get_latest_model.return_value = None
        retval = self.restservice.check_initial_model()
        self.assertFalse(retval)

        self.restservice.statestore.get_latest_model.return_value = ''
        retval = self.restservice.check_initial_model()
        self.assertFalse(retval)


class TestToken(unittest.TestCase):

    @patch('fedn.clients.reducer.control.ReducerControl')
    def setUp(self, mock_control):
        CONFIG = {
            'discover_host': 'TEST_HOST',
            'name': 'TEST_NAME',
            'discover_port': 1111,
            'token': True,
            'remote_compute_context': True
        }

        self.restservice = ReducerRestService(CONFIG, mock_control, None)

    def test_encode_decode_auth_token(self):
        SECRET_KEY = 'test_secret'
        token = self.restservice.encode_auth_token(SECRET_KEY)
        payload_success = self.restservice.decode_auth_token(token, SECRET_KEY)
        payload_invalid = self.restservice.decode_auth_token(
            'wrong_token', SECRET_KEY)
        payload_error = self.restservice.decode_auth_token(token, 'wrong_key')

        self.assertEqual(payload_success, "Success")
        self.assertEqual(payload_invalid, "Invalid token.")
        self.assertEqual(payload_error, "Invalid token.")


if __name__ == '__main__':
    unittest.main()



from collections import OrderedDict

import numpy as np

from .helperbase import HelperBase


class Helper(HelperBase):
    """ FEDn helper class for pytorch. """

    def __init__(self):
        """ Initialize helper. """
        super().__init__()
        self.name = "pytorchhelper"

    def increment_average(self, model, model_next, num_examples, total_examples):
        """ Update a weighted incremental average of model weights.

        :param model: Current model weights with keys from torch state_dict.
        :type model: OrderedDict
        :param model_next: New model weights with keys from torch state_dict.
        :type model_next: OrderedDict
        :param num_examples: Number of examples in new model.
        :type num_examples: int
        :param total_examples: Total number of examples.
        :type total_examples: int
        :return: Incremental weighted average of model weights.
        :rtype: OrderedDict
        """
        w = OrderedDict()
        for name in model.keys():
            tensorDiff = model_next[name] - model[name]
            w[name] = model[name] + num_examples*tensorDiff / total_examples
        return w

    def save(self, model, path=None):
        """ Serialize weights to file. The serialized model must be a single binary object.

        :param model: Weights of model with keys from torch state_dict.
        :type model: OrderedDict
        :param path: File path.
        :type path: str
        :return: Path to file (generated as tmp file unless path is set).
        :rtype: str
        """
        if not path:
            path = self.get_tmp_path()
        np.savez_compressed(path, **model)
        return path

    def load(self, path):
        """ Load weights from file or filelike.

        :param path: file path, filehandle, filelike.
        :type path: str
        :return: Weights of model with keys from torch state_dict.
        :rtype: OrderedDict
        """
        a = np.load(path)
        weights_np = OrderedDict()
        for i in a.files:
            weights_np[i] = a[i]
        return weights_np
